/*
 *  Copyright (C) 2015-2020, Thomas Maier-Komor
 *
 *  This is the source code of hostscope.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <assert.h>
#include <errno.h>
#ifndef __CYGWIN__
#include <execinfo.h>
#endif
#include <fcntl.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <unistd.h>

#include "config.h"
#include "common.h"
#include "log.h"
#include "version.h"


typedef enum { off, on, invalid } flag_t;

const char *Copyright = "Copyright 2015-2020, Thomas Maier-Komor";
const char *License = "GPLv3 - see file LICENSE for details";
const char *Version = VERSION;
const char *McastAddr = MCAST_ADDR;
const char *Database = 0;
unsigned Port = MCAST_PORT;
bool ReadDisks = true, ReadNets = true, ReadProcs = true, Daemon = true, ReadStatm = true;
bool ReadProcIO = true, ReadEuid = false, SendProcInfo = true;

int (*Endfunc)() = 0;


static int isEmpty(const char *l)
{
	while (*l) {
		if ((*l != ' ') && (*l != '\t'))
			return 0;
		++l;
	}
	return 1;
}


#ifdef __linux
void sigSegVhandler(int sig)
{
	assert((sig == SIGSEGV) || (sig == SIGABRT));
	void *bt[128];
	int n = backtrace(bt,sizeof(bt)/sizeof(void*));
	char buf[4096];
	char *a;
	if (sig == SIGABRT)
		a = stpcpy(buf,"SIGABRT occured at:\n");
	else
		a = stpcpy(buf,"SIGSEGV occured at:\n");
	int w = write(Log != -1 ? Log : STDERR_FILENO,buf,a-buf);
	if (w == -1)
		close(Log);
	if (Endfunc)
		Endfunc();
	backtrace_symbols_fd(bt,n,Log != -1 ? Log : STDERR_FILENO);
}
#endif


void print_version()
{
	printf("HostScope Version %s\n%s\nLicense: %s\n\n"
		"This program is free software: you can redistribute it and/or modify\n"
		"it under the terms of the GNU General Public License as published by\n"
		"the Free Software Foundation, either version 3 of the License, or\n"
		"(at your option) any later version.\n"
		"\n"
		"This program is distributed in the hope that it will be useful,\n"
		"but WITHOUT ANY WARRANTY; without even the implied warranty of\n"
		"MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n"
		"GNU General Public License for more details.\n"
		"\n"
		"You should have received a copy of the GNU General Public License\n"
		"along with this program.  If not, see <http://www.gnu.org/licenses/>. \n"
		, Version, Copyright, License); 
}


void kill_daemon(const char *lockfile)
{
	dbug("reading lockfile %s\n",lockfile);
	FILE *pidf = fopen(lockfile,"r");
	if (0 == pidf)
		fatal("lock file %s does not exist - no daemon running?\n",lockfile);
	long pid;
	if (fscanf(pidf,"%ld",&pid) != 1)
		fatal("lock file %s seems to be corrupt - pid not found\n",lockfile);
	fclose(pidf);
	info("killing pid %u\n",pid);
	if (-1 == kill(pid,SIGTERM)) {
		if (errno == ESRCH) {
			if (-1 == unlink(lockfile))
				warning("unable to remove lock file %s: %s\n",lockfile,strerror(errno));
		}
		fatal("unable to kill daemon with pid %d: %s\n",pid,strerror(errno));
	}
	if (-1 == unlink(lockfile))
		warning("unable to remove lock file %s: %s\n",lockfile,strerror(errno));
	exit(EXIT_SUCCESS);
}


void setMulticastPort(char *a)
{
	long p = strtol(a,0,0);
	if (p &~0xffff)
		fatal("unable to set port to %d: value out of range\n",p);
	Port = p;
	info("multicast group port set to %u\n",Port);
}


void setMulticastAddr(char *a)
{
	char *c = strrchr(a,':');
	if (c) {
		*c = 0;
		setMulticastPort(++c);
	}
	McastAddr = strdup(a);
	info("setting multicast group address to %s\n",a);
}


static void parseFlag(const char *valuestr, bool *flag, const char *name)
{
	if ((strcasecmp(valuestr,"yes") == 0) || (strcasecmp(valuestr,"on") == 0) || (strcmp(valuestr,"1") == 0) || (strcmp(valuestr,"true") == 0)) {
		*flag = true;
		dbug("%s = true\n",name);
	} else if ((strcasecmp(valuestr,"no") == 0) || (strcasecmp(valuestr,"off") == 0) || (strcmp(valuestr,"0") == 0) || (strcmp(valuestr,"false") == 0)) {
		*flag = false;
		dbug("%s = false\n",name);
	} else {
		warning("invalid argument '%s' for option %s",valuestr,name);
	}
}


static void readConfigFile(const char *cfname)
{
        int df,lineno = 0;
	struct stat st;
	char *cfdata, *line;

	df = open(cfname,O_RDONLY);
	if (df == -1) {
		if (errno == ENOENT)
			info("no config file %s\n",cfname);
		else
			warning("error opening config file %s: %s\n",cfname,strerror(errno));
		return;
	}
	if (-1 == fstat(df,&st)) {
		warning("unable to stat config file %s: %s\n",cfname,strerror(errno));
		close(df);
		return;
	}
	if ((getuid() != st.st_uid) && (st.st_uid != 0)) {
		warning("ignoring config file '%s' from different user\n",cfname);
		close(df);
		return;
	}
	info("reading config file %s\n",cfname);
	cfdata = (char*)malloc(st.st_size+1);
	if (cfdata == 0)
		fatal("out of memory\n");
	int n = read(df,cfdata,st.st_size);
	const int save_errno = errno;
	close(df);
	if (n < 0) {
		warning("error reading %s: %s\n",cfname,strerror(save_errno));
		free(cfdata);
		return;
	}
	cfdata[n] = 0;
	line = cfdata;
	while (line && *line) {
		char key[64],valuestr[64];
		int a;
		++lineno;
		char *nl = strchr(line,'\n');
		if (nl) {
			*nl = 0;
			++nl;
		}
		char *pound = strchr(line,'#');
		if (pound)
			*pound = 0;
		if (isEmpty(line)) {
			line = nl;
			continue;
		}
		a = sscanf(line,"%63[A-Za-z]%*[ \t=:]%63[0-9a-zA-Z.:]",key,valuestr);
		if (a != 2) {
			warning("config file %s, line %d: error parsing '%s'\n",cfname,lineno,line);
			line = nl;
			continue;
		}
		line = nl;
		dbug("parsing key/value pair %s=%s\n",key,valuestr);
		if (strcasecmp(key,"database") == 0) {
			Database = strdup(valuestr);
			info("setting influx database to %s",Database);
		} else if (strcasecmp(key,"mcast") == 0) {
			setMulticastAddr(valuestr);
		} else if (strcasecmp(key,"port") == 0) {
			setMulticastPort(valuestr);
		} else if (strcasecmp(key,"iostat") == 0) {
			parseFlag(valuestr,&ReadProcIO,"iostat");
		} else if (strcasecmp(key,"diskstat") == 0) {
			parseFlag(valuestr,&ReadDisks,"diskstat");
		} else if (strcasecmp(key,"memstat") == 0) {
			parseFlag(valuestr,&ReadStatm,"memstat");
		} else if (strcasecmp(key,"netstat") == 0) {
			parseFlag(valuestr,&ReadNets,"netstat");
		} else if (strcasecmp(key,"procstat") == 0) {
			parseFlag(valuestr,&ReadProcs,"procstat");
		} else if (strcasecmp(key,"euid") == 0) {
			parseFlag(valuestr,&ReadEuid,"euid");
		} else if (strcasecmp(key,"verbose") == 0) {
			setVerbose(valuestr);
		} else if (strcasecmp(key,"influx") == 0) {
			setInfluxAddr(valuestr);
		} else {
			warning("ignoring invalid parameter '%s'\n",key);
		}
	}
}


void initSettings(int argc, char *argv[])
{
	for (int i = 1; i < argc; ++i) {
		if (0 == strcmp("-v",argv[i])) {
			++i;
			setVerbose(argv[i]);
		} else if (0 == strncmp("-v",argv[i],2)) {
			setVerbose(argv[i]+2);
		}
	}
	readConfigFile(PREFIX "/etc/hostscope.conf");
	readConfigFile("/etc/hostscope.conf");
}


static int daemon_running(const char *lockfile)
{
	int pidf = open(lockfile,O_RDONLY,0666);
	if (-1 == pidf) {
		if (errno == EEXIST)
			info("lock file %s does not exist\n",lockfile);
		else
			error("error opening lockfile %s: %s\n",lockfile,strerror(errno));
		return 0;
	}
	char buf[64];
	int n = read(pidf,buf,sizeof(buf)-1);
	if (n == -1) {
		error("unable to read existing lock file %s: %s\n",lockfile,strerror(errno));
		close(pidf);
		return 1;
	}
	if (n == 0) {
		sleep(1);
		n = pread(pidf,buf,sizeof(buf)-1,0);
		if (n == 0) {
			warning("removing empty stale pid file\n");
			close(pidf);
			int r = unlink(lockfile);
			if (r == -1)
				error("unable to remove stale and empty pid file: %s\n",strerror(errno));
			return 0;
		}
	}
	close(pidf);
	buf[n] = 0;
	long xpid = strtol(buf,0,0);
	assert(xpid != 0);
	snprintf(buf,sizeof(buf),"/proc/%ld",xpid);
	struct stat st;
	if (-1 == stat(buf,&st)) {
		// pid does not exist
		info("no daemon running\n");
		warning("removing stale lock file\n");
		if (-1 == unlink(lockfile))
			error("unable to remove %s: %s\n",lockfile,strerror(errno));
		return 0;
	}
	info("daemon with pid %ld exists\n",xpid);
	return 1;
}


int create_lockfile(const char *lockfile)
{
	int pidf = open(lockfile,O_RDWR|O_TRUNC|O_CREAT|O_EXCL,0666);
	if (pidf != -1)
		return pidf;
	if (errno != EEXIST)
		fatal("unable to create lock file %s: %s\n",strerror(errno));
	if (daemon_running(lockfile))
		fatal("daemon already running\n");
	pidf = open(lockfile,O_RDWR|O_TRUNC|O_CREAT|O_EXCL,0666);
	if (pidf == -1)
		fatal("unable to recreate lock file %s: %s\n",strerror(errno));
	return pidf;
}


void start_daemon(const char *lockfile)
{
	int pidf = create_lockfile(lockfile);
	int pid = fork();
	if (-1 == pid)
		fatal("daemon creation failed while forking: %s\n",strerror(errno));
	if (pid != 0)
		exit(EXIT_SUCCESS);
	int cpid = fork();
	if (-1 == cpid)
		fatal("daemon creation failed while forking: %s\n",strerror(errno));
	if (cpid == 0) {
		close(pidf);
		return;
	}
	char buf[16];
	size_t l = snprintf(buf,sizeof(buf),"%u\n",cpid);
	int n = write(pidf,buf,l);
	assert(n != -1);
	(void) close(pidf);
	exit(EXIT_SUCCESS);
}


