/*
 *  Copyright (C) 2015-2021, Thomas Maier-Komor
 *
 *  This is the source code of hostscope.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <arpa/inet.h>
#include <errno.h>
#include <fcntl.h>
#include <libgen.h>
#include <ncurses.h>
#include <netinet/in.h>
#include <poll.h>
#include <signal.h>
#include <stdlib.h>
#include <strings.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <unistd.h>

#include <iostream>
#include <sstream>

#include "config.h"
#include "common.h"
#include "log.h"
#include "support.h"

#ifdef WFC
#include "hostscope_wfc.h"
#else
#include "hostscope.pb.h"
#endif

#define B_BLACK 1
#define T_WHITE 1
#define B_WHITE 2
#define B_RED 3
#define B_GREEN 4
#define B_BLUE 5
#define B_YELLOW 6
#define B_CYAN 7
#define B_MAGENTA 8
#define T_GREEN 9
#define T_RED 10
#define T_YELLOW 11
#define T_CYAN 12

#define LOCKFILE RUNDIR "/hostscoped.pid"

using namespace std;


static unsigned Rows, Cols, Localmode = 0, MinUid = DEFAULT_MINUSERUID,
		FirstUserUid = DEFAULT_MINUSERUID;
static bool  Sigwinch = false, LoadAv = true, DisplayHelp = false;
static char Packet[PACKETSIZE];	/* maximum size of one packet */

static vector<HostScope*> Hosts;
static vector<time_t> RTime;
static unsigned Host = 0;
static WINDOW *W = 0;


static void
sighandler(int sig)
{
	if (sig == SIGINT || sig == SIGTERM)
		exit(EXIT_FAILURE);
	assert(sig == SIGWINCH);
	Sigwinch = true;
}


static const char *
daemon_binary(char *bin, size_t len)
{
	snprintf(bin,len,"/proc/%d/exe",getpid());
	char tmp[PATH_MAX];
	int n = readlink(bin,tmp,len);
	assert(n > 0);
	strcpy(bin,tmp);
	while (bin[--n] != '/');
	++n;
	strcpy(stpcpy(bin+n,DAEMONDIR),"/" DAEMONEXE);
	dbug("daemonexe is %s\n",bin);
	if (access(bin,X_OK) != 0) {
		error("cannot find hostscoped at %s\n",bin);
		return 0;
	}
	return bin;
}


static int setup_source()
{
	if (Localmode == 0) {
		struct sockaddr_in MA;
		struct ip_mreq mreq;
		int on = 1, fl;
		int sock;

		sock = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
		if (-1 == sock)
			error("cannot create socket: %s\n",strerror(errno));
		fl = fcntl(sock,F_GETFL,0);
		if (-1 == fcntl(sock,F_SETFL,fl | O_NONBLOCK)) 
			warning("could not set socket to nonblocking: %s\n",strerror(errno));
		if (-1 == setsockopt(sock,SOL_SOCKET,SO_REUSEADDR,&on,sizeof(on)))
			error("cannot set socket to reuse address: %s\n",strerror(errno));
		bzero(&MA,sizeof(MA));
		MA.sin_family = AF_INET;
		MA.sin_port = htons(Port);
#if defined __CYGWIN__
		MA.sin_addr.s_addr = htonl(INADDR_ANY);
		mreq.imr_interface.s_addr = htonl(INADDR_ANY);
		mreq.imr_multiaddr.s_addr = inet_addr(McastAddr);
#else
		mreq.imr_interface.s_addr = htonl(INADDR_ANY);
		if (-1 == inet_pton(AF_INET, McastAddr, &MA.sin_addr))
			error("cannot set multicast address: %s\n",strerror(errno));
		memcpy(&mreq.imr_multiaddr.s_addr,(void *)&MA.sin_addr,sizeof(struct in_addr));
#endif
		if (-1 == bind(sock,(struct sockaddr *)&MA,sizeof(MA))) {
			warning("cannot bind socket: %s\n",strerror(errno));
		} else if (-1 == setsockopt(sock,IPPROTO_IP,IP_ADD_MEMBERSHIP,&mreq,sizeof(mreq))) {
			warning("could not join multicast group: %s\n",strerror(errno));
		} else
			return sock;
		close(sock);
		info("falling back to local mode\n");
		Localmode = 1;
	} 
	// no else clause, because this is the fallback mode triggered above
	if (Localmode) {
		char bin[PATH_MAX];
		if (0 == daemon_binary(bin,sizeof(bin)))
			return -1;
		strcat(bin," -L");
		dbug("attaching to %s\n",bin);
		FILE *in = popen(bin,"r");
		if (in == 0)
			error("could not attach hostscoped locally: %s\n",strerror(errno));
		int fid = fileno(in);
		int err = fcntl(fid,F_SETFL,O_NONBLOCK);
		if (-1 == err)
			warning("could not set pipe to nonblocking: %s\n",strerror(errno));
		dbug("successfully attached to hostscoped via stream %p (fileno %d)\n",in,fid);
		return fid;
	}
	return -1;
}


static int receive_stats(int sock)
{
	int num;

	if (Localmode) {
		num = read(sock,Packet,sizeof(Packet));
	} else {
#if defined(__CYGWIN__)
		num = recvfrom(sock,Packet,sizeof(Packet),0,0,0);
#else
		struct sockaddr_in from;
		socklen_t len = sizeof(from);
		num = recvfrom(sock,Packet,sizeof(Packet),0,(struct sockaddr *)&from,&len);
#endif
	}
	if (num == -1) {
		if (errno == EWOULDBLOCK)
			return 0;
		error("cannot read from socket: %s\n",strerror(errno));
		return 0;
	}
	dbug("received %d bytes\n",num);
	return num;
}


static unsigned print_header(const HostScope &t)
{
	TimeStat(__FUNCTION__);
	const char *hn = "", *on = "", *ov = 0, *mn = "";
	assert(t.has_hostinfo());
	const HostScope_HostInfo &hi = t.hostinfo();
	if (hi.has_minuseruid())
		FirstUserUid = hi.minuseruid();
	else
		FirstUserUid = DEFAULT_MINUSERUID;
	if (hi.has_hostname())
		hn = hi.hostname().c_str();
	if (hi.has_osname())
		on = hi.osname().c_str();
	if (hi.has_machtype())
		mn = hi.machtype().c_str();
	if (hi.has_osrelease())
		ov = hi.osrelease().c_str();
	attron(COLOR_PAIR(B_BLACK));

	char line[LINELEN], *l = line;
	if (Cols > sizeof(line))
		Cols = sizeof(line);
	const HostScope_MetaData &md = t.metadata();
	if (md.protocol_version() != PROTOCOL_VERSION)
		warnonce("hostscope server has protocol version %u, but client %u - please upgrade\n",md.protocol_version(),PROTOCOL_VERSION);
	if (strcmp(md.hostscope_version().c_str(),Version))
		warnonce("upgrade version: %s's server %s, client %s\n",hn,md.hostscope_version().c_str(),Version);
	l += snprintf(l,sizeof(line)-(l-line),"%s %s/%s ",hn,on,mn);
	int nps = t.power_supplies_size();
	if (nps > 0)
		*l++ = '[';
	for (int i = 0; i < nps; ++i) {
		const HostScope_PowerSupply &p = t.power_supplies(i);
		l = stpcpy(l,p.name().c_str());
		*l++ = ' ';
		if (p.has_capacity()) {
			unsigned cap = p.capacity();
			if (cap < 20) { 
				*l = 0;
				addstr(line);
				attron(cap < 10 ? COLOR_PAIR(T_RED) : COLOR_PAIR(T_YELLOW));
				printw("%u%%",cap);
				attron(COLOR_PAIR(T_WHITE));
				l = line;
			} else {
				l += sprintf(l,"%u%%",cap);
			}
		} else {
			if (p.state() == "off") {
				*l = 0;
				addstr(line);
				attron(COLOR_PAIR(T_YELLOW));
				addstr(p.state().c_str());
				attron(COLOR_PAIR(T_WHITE));
				l = line;
			} else {
				l = stpcpy(l,p.state().c_str());
			}
		}
		if (i == (nps - 1))
			*l++ = ']';
		else
			l = stpcpy(l,", ");
	}
	if ((nps == 0) && (ov != 0))
		l = stpcpy(l,ov);
	l = stpcpy(l,": up ");

	clock_t ut = hi.uptime();
	int d,h,m,b = ut;
	d = b / (60*60*24);
	b -= d * (60*60*24);
	h = b / (60*60);
	b -= h * (60*60);
	m = b / (60);
	b -= m * (60);
	if (d == 1) {
		l = stpcpy(l,"1 day ");
	} else if (d) {
		l += snprintf(l,sizeof(line)-(l-line),"%d days ",d);
	}
	l += snprintf(l,sizeof(line)-(l-line),"%d:%02d:%02d",h,m,b);
	addstr(line);

	time_t ts = md.timestamp();
	char timestr[20];
	struct tm tm;
	localtime_r(&ts,&tm);
	int tl = strftime(timestr,sizeof(timestr),"%T",&tm);
	mvaddstr(0,Cols-tl,timestr);
	hline(ACS_HLINE,Cols);
	return 1;
}


static unsigned color_bar(unsigned steps, double d, unsigned c)
{
	d /= 100.0;
	unsigned n = (unsigned)nearbyint((double)steps*d/100);
	if (n == 0)
		return 0;
	attron(c);
	if (n == 1) {
		if (d < 10)
			printw("%1.0f",d);
		else
			printw(" ");
		return 1;
	}
	if (n == 2) {
		printw("%2.0f",d);
		return 2;
	}
	char format[16];
	sprintf(format,"%%%u.0f",n);
	printw(format,d);
	return n;
}


void print_cpu(const HostScope_LinuxCPU &cpu, int id)
{
	char line[LINELEN], *l = line;
	bool red = false;
	if (id != -1)
		l += snprintf(l,sizeof(line)-(l-line),"%3d ",id);
	if (cpu.has_currmhz())
		l += sprintf(l,"%4u ",cpu.currmhz());
	else
		l += snprintf(l,sizeof(line)-(l-line),"     ");
	if (cpu.has_crittemp() && cpu.has_currtemp() && cpu.currtemp() >= cpu.crittemp()) {
		addstr(line);
		attron(COLOR_PAIR(T_RED));
		printw("%3.0f ",(float)cpu.currtemp()/10);
		l = line;
		red = true;
	} else if (cpu.has_currtemp()) {
		l += snprintf(l,sizeof(line)-(l-line),"%3.0f ",(float)cpu.currtemp()/10);
	} else {
		l += snprintf(l,sizeof(line)-(l-line),"    ");
	}
	l = word2perc(cpu.user(),l);
	*l = ' ';
	l = word2perc(cpu.system(),++l);
	*l = ' ';
	l = word2perc(cpu.nice(),++l);
	*l = ' ';
	l = word2perc(cpu.iowait(),++l);
	*l = ' ';
	l = word2perc(cpu.irq(),++l);
	*l = ' ';
	l = word2perc(cpu.vhost(),++l);
	*l = ' ';
	l = word2perc(cpu.idle(),++l);
	*l++ = ' ';
	*l++ = ' ';
	*l = 0;
	if (red)
		attron(COLOR_PAIR(T_WHITE));
	addstr(line);
	int steps = Cols - strlen(line);
	unsigned idle = steps;
	idle -= color_bar(steps,cpu.user(),COLOR_PAIR(B_GREEN));
	idle -= color_bar(steps,cpu.system(),COLOR_PAIR(B_YELLOW));
	idle -= color_bar(steps,cpu.nice(),COLOR_PAIR(B_BLUE));
	idle -= color_bar(steps,cpu.iowait(),COLOR_PAIR(B_MAGENTA));
	idle -= color_bar(steps,cpu.irq(),COLOR_PAIR(B_RED));
	idle -= color_bar(steps,cpu.vhost(),COLOR_PAIR(B_CYAN));
	attron(COLOR_PAIR(B_BLACK));
}


static unsigned print_cpu_summary(unsigned ln, const HostScope &t)
{
	HostScope_LinuxCPU low = t.linux_cpus(0), high = low;
	unsigned nidle = 0, nbusy = 0;
	unsigned ua = 0, na = 0, sa = 0, ia = 0, wa = 0, ta = 0, qa = 0, va = 0;
	unsigned ma = 0, tc = 0;
	int32_t lt = INT32_MAX, ht = INT32_MIN;
	for (const auto &cpu : t.linux_cpus()) {
		unsigned idle = cpu.idle();
		if (idle < high.idle())
			high = cpu;
		else if (idle > low.idle())
			low = cpu;
		ia += idle;
		if (idle < 100)
			++nbusy;
		else if (idle > 9000)
			++nidle;
		ua += cpu.user();
		na += cpu.nice();
		sa += cpu.system();
		wa += cpu.iowait();
		qa += cpu.irq();
		va += cpu.vhost();
	}
	for (const auto &cpu : t.linux_cpus()) {
		unsigned mhz = cpu.currmhz();
		ma += mhz;
		if (low.currmhz() > mhz)
			low.set_currmhz(mhz);
		if (high.currmhz() < mhz)
			high.set_currmhz(mhz);
		if (cpu.has_currtemp()) {
			++tc;
			int32_t temp = cpu.currtemp();
			if (lt > temp)
				lt = temp;
			else if (ht < temp)
				ht = temp;
			ta += temp;
		}
	}
	HostScope_LinuxCPU avg;
	unsigned ncpu = t.linux_cpus_size();
	avg.set_idle(ia/ncpu);
	avg.set_user(ua/ncpu);
	avg.set_nice(na/ncpu);
	avg.set_system(sa/ncpu);
	avg.set_iowait(wa/ncpu);
	avg.set_irq(qa/ncpu);
	avg.set_vhost(va/ncpu);
	avg.set_currmhz(ma/ncpu);
	if (tc > 0) {
		avg.set_currtemp(ta/tc);
		low.set_currtemp(lt);
		high.set_currtemp(ht);
	}
	char line[80];
	snprintf(line,sizeof(line)," (%u Cores: %u idle, %u active, %u busy)",ncpu,nidle,ncpu-nidle-nbusy,nbusy);
	addstr(line);
	mvaddstr(++ln,0,"max ");
	print_cpu(high,-1);
	mvaddstr(++ln,0,"avg ");
	print_cpu(avg,-1);
	mvaddstr(++ln,0,"min ");
	print_cpu(low,-1);
	mvhline(++ln,0,ACS_HLINE,Cols);
	return ln;
}


static unsigned print_cpus(unsigned ln, const HostScope &t)
{
	TimeStat(__FUNCTION__);
	unsigned ncpu = t.linux_cpus_size();
	if (ncpu == 0)
		return ln;
	attron(A_UNDERLINE);
	mvaddstr(++ln,0,"CPU  MHz  \302\260C usr sys nic wai int vir idl"); 
	attroff(A_UNDERLINE);
	if (ncpu > 4)
		return print_cpu_summary(ln,t);
	addstr("  ");
	attron(COLOR_PAIR(B_GREEN));
	addstr("user");
	attron(COLOR_PAIR(T_WHITE));
	addch(' ');
	attron(COLOR_PAIR(B_YELLOW));
	addstr("sys");
	attron(COLOR_PAIR(T_WHITE));
	addch(' ');
	attron(COLOR_PAIR(B_BLUE));
	addstr("nice");
	attron(COLOR_PAIR(T_WHITE));
	addch(' ');
	attron(COLOR_PAIR(B_MAGENTA));
	addstr("wait");
	attron(COLOR_PAIR(T_WHITE));
	addch(' ');
	attron(COLOR_PAIR(B_RED));
	addstr("intr");
	attron(COLOR_PAIR(T_WHITE));
	addch(' ');
	attron(COLOR_PAIR(B_CYAN));
	addstr("virt");
	attron(COLOR_PAIR(T_WHITE));
	for (unsigned i = 0; i < ncpu; ++i) {
		mvaddstr(++ln,0,"");
		print_cpu(t.linux_cpus(i),i);
	}
	mvhline(++ln,0,ACS_HLINE,Cols);
	return ln;
}


static unsigned print_disks(unsigned ln, const HostScope &t)
{
	TimeStat(__FUNCTION__);
	char line[LINELEN];
	int ndisks = t.disks_size();
	if (ndisks == 0)
		return ln;
	attron(A_UNDERLINE);
	mvaddstr(++ln,0,"storage      rd/s  wr/s    Brd/s  Bwr/s   size   opq   busy");
	attroff(A_UNDERLINE);
	for (int i = 0; i < ndisks; ++i) {
		const HostScope_Disk &disk = t.disks(i);
		char *l = line;
		l += sprintf(line,"%-12s ",disk.devname().c_str());
		if (disk.has_nread())
			l = dbl2str_r(disk.nread()*1E-4,l);
		else
			l = stpcpy(l,"     ");
		*l++ = ' ';
		if (disk.has_nwrite())
			l = dbl2str_r(disk.nwrite()*1E-4,l);
		else
			l = stpcpy(l,"     ");
		*l++ = ' ';
		*l++ = ' ';
		if (disk.has_kread())
			l = kb2str_r(disk.kread(),l);
		else
			l = stpcpy(l," <?>ki");
		*l++ = ' ';
		if (disk.has_kwrite())
			l = kb2str_r(disk.kwrite(),l);
		else
			l = stpcpy(l," <?>ki");
		*l++ = ' ';
		*l++ = ' ';
		if (disk.has_ktotal())
			l = kb2str_r(disk.ktotal(),l);
		else
			l = stpcpy(l," <?>ki");
		*l++ = ' ';
		if (disk.has_opsqueued())
			l = dbl2str_r((double)disk.opsqueued(),l);
		else
			l = stpcpy(l,"     ");
		*l++ = ' ';
		if (disk.has_perc_busy())
			l = word2perc(disk.perc_busy(),l);
		else
			l = stpcpy(l," <?>");
		*l++ = '%';
		*l++ = ' ';
		*l = 0;

		mvaddstr(++ln,0,line);
		unsigned steps = Cols - strlen(line) - 1;
		color_bar(steps,disk.perc_busy(),COLOR_PAIR(B_RED));
		attron(COLOR_PAIR(T_WHITE));
	}
	mvhline(++ln,0,ACS_HLINE,Cols);
	return ln;
}


static unsigned print_networks(unsigned ln, const HostScope &t)
{
	TimeStat(__FUNCTION__);
	char line[LINELEN];
	int ndevs = t.netdevices_size();
	if (ndevs == 0)
		return ln;
	attron(A_UNDERLINE);
	mvaddstr(++ln,0,"networks     rx/s  tx/s    Brx/s  Btx/s  drop  errs");
	attroff(A_UNDERLINE);
	for (int i = 0; i < ndevs; ++i) {
		const HostScope_NetDevice &dev = t.netdevices(i);
		char *l = line;
		l += sprintf(line,"%-12s ",dev.devname().c_str());
		dbug("net %s: %u %u %u %u\n",dev.devname().c_str(),dev.nreceived(),dev.ntransmit(),dev.kreceived(),dev.ktransmit());
		if (dev.has_nreceived())
			l = dbl2str_r((double)dev.nreceived()*1E-4,l);
		else
			l = stpcpy(l,"     ");
		*l++ = ' ';
		if (dev.has_ntransmit())
			l = dbl2str_r((double)dev.ntransmit()*1E-4,l);
		else
			l = stpcpy(l,"     ");
		*l++ = ' ';
		*l++ = ' ';
		if (dev.has_kreceived()) {
			l = kb2str_r((double)dev.kreceived()*1E-4,l);
			*l++ = ' ';
		} else
			l = stpcpy(l,"     ");
		if (dev.has_ktransmit()) {
			l = kb2str_r((double)dev.ktransmit()*1E-4,l);
			*l++ = ' ';
		} else
			l = stpcpy(l,"     ");
		if (dev.has_dtransmit() & dev.has_dreceived())
			l = dbl2str_r((double)(dev.dtransmit()+dev.dreceived())*1E-4,l);
		*l++ = ' ';
		if (dev.has_etransmit() & dev.has_ereceived())
			l = dbl2str_r((double)(dev.etransmit()+dev.ereceived()),l);
		*l = 0;
		mvaddstr(++ln,0,line);
	}
	mvhline(++ln,0,ACS_HLINE,Cols);
	return ln;
}


static unsigned print_mem(unsigned ln, const HostScope &t)
{
	TimeStat(__FUNCTION__);
	if (!t.has_mem()) 
		return ln;
	char line[LINELEN*4];
	const HostScope_Memory &mem = t.mem();
	const HostScope_LinuxVmstat &v = t.linux_vmstat();
	uint64_t memtotal = mem.physical();
	uint64_t memfree;
	if (v.has_nr_free_pages())
		memfree = v.nr_free_pages()<<2;
	else
		memfree = mem.free();	// fallback - vmstat info used if available
	uint64_t swapsize = mem.swapsize();
	uint64_t pagedout = mem.pagedout();
	//uint64_t locked = mem.locked();	-- vmstat info used
	const HostScope_LoadAvg &la = t.loadavg();
	double l1 = ((double)la.loadav1m())/1000.0;
	double l5 = ((double)la.loadav5m())/1000.0;
	double l15 = ((double)la.loadav15m())/1000.0;

	const HostScope_LinuxInfo &li = t.linux_info();
//	uint64_t locked = v.nr_mlock()<<2;	-- unused

	attron(A_UNDERLINE);
	if (LoadAv)
		mvaddstr(++ln,0,"memory     [B]   pages      [B]   paging   [B/s]   system [1/s]    load avg.");
	else
		mvaddstr(++ln,0,"memory     [B]   pages      [B]   paging   [B/s]   system [1/s]");
	attroff(A_UNDERLINE);

	char *l = stpcpy(line,"total   ");
	l = kb2str_r(memtotal,l);
	l = stpcpy(l,"   anon    ");
	if (v.has_nr_anon_pages())
		l = kb2str_r(v.nr_anon_pages()<<2,l);
	else
		l = stpcpy(l," unav.");
	l = stpcpy(l,"   pgpgin  ");
	if (v.has_pgpgin())
		l = kb2str_r(v.pgpgin()<<2,l);
	else
		l = stpcpy(l," unav.");
	l = stpcpy(l,"   fork    ");
	if (li.has_forks())
		l = dbl2str_r((double)li.forks()*1E-4,l);
	else
		l = stpcpy(l,"unav.");
	if (LoadAv)
		l += snprintf(l,sizeof(line)-(l-line),"    1m %3.2f",l1);

	l = stpcpy(l,"\nfree    ");
	l = kb2str_r(memfree,l);
	l = stpcpy(l,"   files   ");
	if (v.has_nr_file_pages())
		l = kb2str_r(v.nr_file_pages()<<2,l);
	else
		l = stpcpy(l," unav.");
	l = stpcpy(l,"   pgpgout ");
	if (v.has_pgpgout())
		l = kb2str_r(v.pgpgout()<<2,l);
	else
		l = stpcpy(l," unav.");
	l = stpcpy(l,"   ctxsw   ");
	if (li.has_ctxtsw())
		l = dbl2str_r((double)li.ctxtsw()*1E-4,l);
	else
		l = stpcpy(l,"unav.");
	if (LoadAv)
		l += snprintf(l,sizeof(line)-(l-line),"    5m %3.2f",l5);

	l = stpcpy(l,"\nswap    ");
	l = kb2str_r(swapsize,l);
	l = stpcpy(l,"   kstack  ");
	if (v.has_nr_kernel_stack())
		l = kb2str_r(v.nr_kernel_stack()<<2,l);
	else
		l = stpcpy(l," unav.");
	l = stpcpy(l,"   pgfree  ");
	if (v.has_pgfree())
		l = kb2str_r(v.pgfree()<<2,l);
	else
		l = stpcpy(l," unav.");
	l = stpcpy(l,"   intr    ");
	if (li.has_intrs())
		l = dbl2str_r((double)li.intrs()*1E-4,l);
	else
		l = stpcpy(l,"unav.");
	if (LoadAv)
		l += snprintf(l,sizeof(line)-(l-line),"   15m %3.2f",l15);

	l = stpcpy(l,"\nswapped ");
	l = kb2str_r(pagedout,l);
	l = stpcpy(l,"   shmem   ");
	if (v.has_nr_shmem())
		l = kb2str_r(v.nr_shmem()<<2,l);
	else
		l = stpcpy(l," unav.");
	l = stpcpy(l,"   pgfault ");
	if (v.has_pgfault())
		l = kb2str_r(v.pgfault()<<2,l);
	else
		l = stpcpy(l," unav.");
	l = stpcpy(l,"   sirq    ");
	if (li.has_sirqs())
		l = dbl2str_r((double)li.sirqs()*1E-4,l);
	else
		l = stpcpy(l,"unav.");
	mvaddstr(++ln,0,line);
	ln += 4;
	mvhline(ln,0,ACS_HLINE,Cols);
	return ln;
}


static unsigned print_arc(unsigned ln, const HostScope &t)
{
	TimeStat(__FUNCTION__);
	if (!t.has_arc())
		return ln;
	const auto &arc = t.arc();
	char line[LINELEN];
	attron(A_UNDERLINE);
	mvaddstr(++ln,0,"ZFS          [B]   [1/s]  fetch   data  mdata   L2     I/O   [B/s]");
	attroff(A_UNDERLINE);

	char *l = line;
	l = stpcpy(l,"ARC size: ");
	if (arc.has_size())
		l = kb2str_r(arc.size(),l);
	else
		l = stpcpy(l," unav.");
	l = stpcpy(l,"  hits  : ");
	if (arc.has_prefetch_hit())
		l = dbl2str_r((double)arc.prefetch_hit(),l);
	else
		l = stpcpy(l," unav.");
	*l++ = ' ';
	if (arc.has_data_hit())
		l = kb2str_r(arc.data_hit(),l);
	else
		l = stpcpy(l," unav.");
	*l++ = ' ';
	if (arc.has_metadata_hit())
		l = kb2str_r(arc.metadata_hit(),l);
	else
		l = stpcpy(l," unav.");
	*l++ = ' ';
	if (arc.has_l2_hit())
		l = dbl2str_r(arc.l2_hit(),l);
	else
		l = stpcpy(l,"unav.");
	l = stpcpy(l,"  read : ");
	if (arc.has_l2_read())
		l = kb2str_r(arc.l2_read(),l);
	else
		l = stpcpy(l," unav.");
	mvaddstr(++ln,0,line);

	l = line;
	l = stpcpy(l,"L2 size : ");
	if (arc.has_l2size())
		l = kb2str_r(arc.l2size(),l);
	else
		l = stpcpy(l," unav.");
	l = stpcpy(l,"  misses: ");
	if (arc.has_prefetch_miss())
		l = dbl2str_r(arc.prefetch_miss(),l);
	else
		l = stpcpy(l," unav.");
	*l++ = ' ';
	if (arc.has_data_miss())
		l = kb2str_r(arc.data_miss(),l);
	else
		l = stpcpy(l," unav.");
	*l++ = ' ';
	if (arc.has_metadata_miss())
		l = kb2str_r(arc.metadata_miss(),l);
	else
		l = stpcpy(l," unav.");
	*l++ = ' ';
	if (arc.has_l2_miss())
		l = dbl2str_r(arc.l2_miss(),l);
	else
		l = stpcpy(l,"unav.");
	l = stpcpy(l,"  write: ");
	if (arc.has_l2_read())
		l = kb2str_r(arc.l2_write(),l);
	else
		l = stpcpy(l," unav.");
	mvaddstr(++ln,0,line);

	mvhline(++ln,0,ACS_HLINE,Cols);
	return ln;
}


static unsigned print_proc(const HostScope_Process &proc, unsigned ln, bool with_shmem)
{
	char line[LINELEN], *l = line;
	l += sprintf(line,"%7d ",proc.pid());
	if (proc.has_nthreads())
		l += sprintf(l,"%4d  ",proc.nthreads());
	else
		l = stpcpy(l,"<?>   ");
	if (proc.has_rssmem())
		l = kb2str_r(proc.rssmem(),l);
	else
		l = stpcpy(l," <?>ki");
	if (!with_shmem);
	else if (proc.has_sharedmem()) {
		l = stpcpy(l," ");
		l = kb2str_r(proc.sharedmem(),l);
	} else
		l = stpcpy(l,"  <?>ki");
	if (proc.has_minflts()) {
		l = stpcpy(l,"  ");
		l = dbl2str_r((double)proc.minflts()*1E-4,l);
	} else
		l = stpcpy(l,"   <?> ");
	if (proc.has_majflts()) {
		l = stpcpy(l,"  ");
		l = dbl2str_r((double)proc.majflts()*1E-4,l);
	} else
		l = stpcpy(l,"   <?> ");
	if (proc.has_putime()) {
		l = stpcpy(l,"  ");
		l = word2perc(proc.putime(),l);
	} else
		l = stpcpy(l,"  <?>%");
	if (proc.has_pstime()) {
		l = stpcpy(l,"  ");
		l = word2perc(proc.pstime(),l);
	} else
		l = stpcpy(l,"  <?>%");
	if (proc.has_rchars()) {
		l = stpcpy(l,"  ");
		l = kb2str_r(proc.rchars(),l);
	} else
		l = stpcpy(l,"   <?>ki");
	if (proc.has_wchars()) {
		l = stpcpy(l,"  ");
		l = kb2str_r(proc.wchars(),l);
	} else
		l = stpcpy(l,"   <?>ki");
	assert(l < line+sizeof(line));
	l += snprintf(l,sizeof(line)-(l-line)-1,"  %s",proc.command().c_str());
	assert(l < line+sizeof(line));
	line[sizeof(line)-1] = 0;
	mvaddstr(++ln,0,line);
	return ln;
}


static unsigned print_procs(unsigned ln, const HostScope &t)
{
	TimeStat(__FUNCTION__);
	int nprocs = t.processes_size();
	if (nprocs == 0)
		return ln;
	bool with_shmem = false;
	for (int i = 0; i < nprocs; ++i) {
		const HostScope_Process &proc = t.processes(i);
		if (proc.has_sharedmem()) {
			with_shmem = true;
			break;
		}
	}
	attron(COLOR_PAIR(B_BLACK)|A_UNDERLINE);
	if (with_shmem)
		mvaddstr(++ln,0,"  pid   nthr   rss   shared minflt majflt  %usr %sys rchar/s wchar/s command");
	else
		mvaddstr(++ln,0,"  pid   nthr   rss   minflt majflt  %usr %sys rchar/s wchar/s command");
	attroff(A_UNDERLINE);
	for (int i = 0; i < nprocs; ++i) {
		const HostScope_Process &proc = t.processes(i);
		ln = print_proc(proc,ln,with_shmem);
	}
	mvhline(++ln,0,ACS_HLINE,Cols);
	return ln;
}


static unsigned print_user(const HostScope_User &user, unsigned ln, bool with_shmem)
{
	char line[LINELEN], *l = line;
	l += sprintf(line,"%5d/",user.nproc());
	if (user.has_nthreads())
		l += sprintf(l,"%4d  ",user.nthreads());
	else
		l = stpcpy(l,"<?>   ");
	if (user.has_rssmem())
		l = kb2str_r(user.rssmem(),l);
	else
		l = stpcpy(l," <?>");
	if (with_shmem) {
		if (user.has_sharedmem()) {
			l = stpcpy(l,"  ");
			l = kb2str_r(user.sharedmem(),l);
		} else
			l = stpcpy(l,"   <?>ki");
	}
	if (user.has_minflts()) {
		l = stpcpy(l,"  ");
		l = dbl2str_r((double)user.minflts()*1E-4,l);
	} else
		l = stpcpy(l,"   <?> ");
	if (user.has_majflts()) {
		l = stpcpy(l,"  ");
		l = dbl2str_r((double)user.majflts()*1E-4,l);
	} else
		l = stpcpy(l,"   <?> ");
	if (user.has_putime()) {
		l = stpcpy(l,"  ");
		l = word2perc(user.putime(),l);
	} else
		l = stpcpy(l,"  <?>%");
	if (user.has_pstime()) {
		l = stpcpy(l,"  ");
		l = word2perc(user.pstime(),l);
	} else
		l = stpcpy(l,"  <?>%");
	if (user.has_rchars()) {
		l = stpcpy(l,"   ");
		l = kb2str_r(user.rchars(),l);
	} else
		l = stpcpy(l,"    <?>ki");
	if (user.has_wchars()) {
		l = stpcpy(l,"  ");
		l = kb2str_r(user.wchars(),l);
	} else
		l = stpcpy(l,"   <?>ki");
	*l++ = ' ';
	*l++ = ' ';
	strcpy(l,user.name().c_str());
	mvaddstr(++ln,0,line);
	return ln;
}


static unsigned print_users(unsigned ln, const HostScope &t)
{
	TimeStat(__FUNCTION__);
	int nusers = t.users_size();
	if (nusers == 0)
		return ln;
	attron(COLOR_PAIR(B_BLACK)|A_UNDERLINE);
	bool with_shmem = false;
	for (int i = 0; i < nusers; ++i) {
		const HostScope_User &user = t.users(i);
		if (user.nproc() == 0)
			continue;
		if (user.has_sharedmem()) {
			with_shmem = true;
			break;
		}
	}
	if (with_shmem)
		mvaddstr(++ln,0,"nproc nthr   rss   shared minflt majflt  %usr %sys rchar/s wchar/s  user");
	else
		mvaddstr(++ln,0,"nproc nthr   rss   minflt majflt  %usr %sys rchar/s wchar/s  user");
	attroff(A_UNDERLINE);
	HostScope_User sys;
	sys.set_name("<system>");

	for (int i = 0; i < nusers; ++i) {
		const HostScope_User &user = t.users(i);
		if (user.nproc() == 0)
			continue;
		if (user.uid() >= MinUid)
			continue;
		sys.set_nproc(sys.nproc()+user.nproc());
		sys.set_nthreads(sys.nthreads()+user.nthreads());
		sys.set_rssmem(sys.rssmem()+user.rssmem());
		sys.set_sharedmem(sys.sharedmem()+user.sharedmem());
		sys.set_minflts(sys.minflts()+user.minflts());
		sys.set_majflts(sys.majflts()+user.majflts());
		sys.set_putime(sys.putime()+user.putime());
		sys.set_pstime(sys.pstime()+user.pstime());
		if (user.has_rchars())
			sys.set_rchars(sys.rchars()+user.rchars());
		if (user.has_wchars())
			sys.set_wchars(sys.wchars()+user.wchars());
	}
	if (sys.nproc() > 0) {
		ln = print_user(sys,ln,with_shmem);
	}
	for (int i = 0; i < nusers; ++i) {
		const HostScope_User &user = t.users(i);
		unsigned nproc = user.nproc();
		if (nproc == 0)
			continue;
		if (user.uid() < MinUid)
			continue;
		ln = print_user(user,ln,with_shmem);
	}
	mvhline(++ln,0,ACS_HLINE,Cols);
	return ln;
}


static void display_help()
{
	mvprintw(0,0,"HostScope Version %s, %s\n"
		"License: %s\n"
		"\n"
		"Keys:\n"
		"<?> : this help screen           <q> : quit hostscope\n"
		"<h> : switch to next host        <H> : switch to previous host\n"
		"<u> : toggle system users view   <l> : toggle load average view\n"
		"\n"
		"CPU metrics:\n"
		"user: user processes              syst: operating sys activities\n"
		"wait: percentage waiting of I/O   nice: nice processes\n"
		"intr: interrupt handling          virt: virtual clients (guests)\n"
		"idle: CPU idle percentag          temp: CPU temperature in Celsius\n"
		"\n"
		"Page metrics:          Paging metrics:             System metrics:\n"
		"anon  : anonymous      pgpgin : pages paged in     fork : process forks\n"
		"file  : files          pgpgout: pages paged out    ctxsw: context switches\n"
		"kstack: kernel stack   pgree  : freed pages        intr : hardware interrupts\n"
		"shmem : shared pages   pgfault: page faults        sirq : software interrupts\n"
		"\n"
		"storage:\n"
		"rd/wr: read/write operations       Brd/Bwr: bytes read/written\n"
		"opq  : operations queued           busy   : %% device busy\n"
		"\n"
		"networks:\n"
		"rx/tx: receive/transmit count      Brx/Btx: bytes received/transmitted\n"
		"drop:  number of packets dropped   errs: nummer of broken packets\n"
		"\n"
		"processes:\n"
		"pid   : process ID     nthr   : number of threads  rss    : resident memory\n"
		"minflt: minor faults   majflt : major faults       %%usr   : user CPU %%\n"
		"%%sys  : system CPU %%   rchar/s: bytes read/sec.    wchar/s: bytes written/sec.\n"
		, Version, Copyright, License
		);
	refresh();
}


static const HostScope &process_input(int sock, int itv)
{
	TimeStat(__FUNCTION__);
	struct pollfd fds[2];
	char c;
	unsigned trys = 0;
	
retry:
	switch (++trys) {
	case 2:
		warnonce("waiting for data from daemon...\n\r");
		break;
	case 5:
		warnonce("still no data received - maybe your firewall blocks multicasts?\n\r"
			"Please verify that \"net.ipv4.all.rp_filter=0\" is in /etc/sysctl.conf\n\r");
		break;
	case 20:
		fatal("still no data - giving up... Please retry in local mode (option -L).\n");
	default:;
	}

	fds[0].fd = STDIN_FILENO;
	fds[0].events = POLLIN;
	fds[1].fd = sock;
	fds[1].events = POLLIN;
	if (-1 == poll(fds,sock == -1 ? 1 : sizeof(fds)/sizeof(struct pollfd),itv * 1000)) {
		if (errno != EINTR)
			error("error polling: %s\n",strerror(errno));
		if (Hosts.empty())
			goto retry;
		return *Hosts[Host];
	}
	if ((fds[0].events & POLLIN) && (1 == read(STDIN_FILENO,&c,sizeof(char)))) {
		unsigned nh = Hosts.size();
		switch (c) {
			case 'h':
				if (++Host >= nh)
					Host = 0;
				break;
			case 'H':
				if (--Host >= nh)
					Host = nh-1;
				break;
			case 'l':
				if (LoadAv)
					LoadAv = false;
				else
					LoadAv = true;
				break;
			case 'q':
				for (unsigned i = 0; i < nh; ++i)
					delete Hosts[i];
				exit(EXIT_SUCCESS);
				break;
			case 'u':
				if (MinUid) 
					MinUid = 0;
				else
					MinUid = FirstUserUid;
				break;
			case '?':
				DisplayHelp ^= true;
				break;
			case ' ':
				if (DisplayHelp)
					DisplayHelp = false;
				break;
			default:
				;
		}
	}
	if (fds[1].events & POLLIN) {
		int n = receive_stats(sock);
		if (n == 0) {
			if (Hosts.empty())
				goto retry;
			return *Hosts[Host];
		}
		static HostScope *t = new HostScope;
		#ifdef WFC
		t->clear();
		int x = t->fromMemory(Packet,n);
		if (x != n) {
			dbug("ignoring broken packet (%d/%u)",x,n);
			goto retry;
		}
		#else
		t->ParseFromArray(Packet,n);
		#endif
		if (Hosts.empty()) {
			Hosts.push_back(t);
			RTime.push_back(time(0));
			t = new HostScope;
		} else {
			const string &newhost = t->hostinfo().hostname();
			int nh = Hosts.size();
			for (int i = 0; i < nh; ++i) {
				if (Hosts[i]->hostinfo().hostname() == newhost) {
					RTime[i] = time(0);
					HostScope *tmp = Hosts[i];
					Hosts[i] = t;
					t = tmp;
					return *Hosts[Host];
				}
			}
			Hosts.push_back(t);
			RTime.push_back(0);	// tag as newly added
			t = new HostScope;
		}
	}
	if (Hosts.empty())
		goto retry;
	return *Hosts[Host];
}


void usage(const char *m, int r)
{
	printf(
	"hostscope [option]\n"
	"Options:\n"
	"-h        : display this help screen and exit\n"
	"-k        : kill hostscope's daemon\n"
	"-L        : operate in local mode (no multicasts)\n"
	"-l <log>  : write log to file <log>\n"
	"-m <a:p>  : set multicast group address and port (optional) to <a:p>\n"
	"-p <port> : set multicast group port to <port>\n"
	"-v <lvl>  : set verbose mode to level <lvl>\n"
	"-V        : display version number and exit\n");
	if (m[0]) 
		printf("\n%s\n",m);
	exit(r);
}


void parse_argv(char a)
{
	switch (a) {
	default:
		abort();
		break;
	case 'h':
		usage("",EXIT_SUCCESS);
		break;
	case 'k':
		{
			char killdaemon[PATH_MAX];
			if (daemon_binary(killdaemon,sizeof(killdaemon))) {
				strcat(killdaemon," -k");
				if (-1 == system(killdaemon))
					fatal("error occured running %s\n");
			}
			exit(EXIT_SUCCESS);
		}
		break;
	case 'L':
		Localmode = true;
		break;
	case 'l':
		setLogfile(optarg);
		break;
	case 'm':
		setMulticastAddr(optarg);
		break;
	case 'p':
		setMulticastPort(optarg);
		break;
	case 'v':
		setVerbose(optarg);
		break;
	case 'V':
		print_version();
		exit(EXIT_SUCCESS);
	case ':':
		usage("missing argument",EXIT_FAILURE);
		break;
	case '?':
		usage("unknown option",EXIT_FAILURE);
		break;
	}
}


static int daemon_running()
{
	int pidf = open(LOCKFILE,O_RDONLY,0);
	if (pidf == -1)
		return 0;
	dbug("daemon_running(): lock file exists\n");
	char buf[64];
	int n = read(pidf,buf,sizeof(buf));
	close(pidf);
	if (n == 0)
		return 0;
	if (n == -1)
		fatal("unable to read existing lock file " LOCKFILE ": %s\n",strerror(errno));
	long xpid = strtol(buf,0,10);
	dbug("existing daemon pid %d\n",xpid);
	assert(xpid != 0);
	snprintf(buf,sizeof(buf),"/proc/%ld/comm",xpid);
	int exefd = open(buf,O_RDONLY);
	if (-1 == exefd)	// pid  does not exists
		return 0;
	n = read(exefd,buf,sizeof(buf));
	if (n < 0) {
		warning("unable to read comm of daemon: %s\n",strerror(errno));
		return 0;
	}
	assert(n > 0);
	close(exefd);
	buf[n] = 0;
	dbug("existing daemon comm(%d): '%s'\n",n,buf);
	if (strcmp(buf,"hostscoped\n")) // daemon already running
		warning("daemon's pid has unexpected comm: %s\n",buf);
	return 1;
}


void setInfluxAddr(char *a)
{

}


int main(int argc, char *argv[])
{
#ifdef GOOGLE_PROTOBUF_VERIFY_VERSION
	GOOGLE_PROTOBUF_VERIFY_VERSION;
#endif
	int i;
#ifdef __linux
	signal(SIGSEGV,sigSegVhandler);
	signal(SIGABRT,sigSegVhandler);
#endif
	initSettings(argc,argv);
	History = 20;
	while ((i = getopt(argc,argv,":hkl:Lm:p:Vv:")) != -1) {
		parse_argv(i);
	}
	if (!Localmode && !daemon_running()) {
		char daemon[PATH_MAX];
		if (daemon_binary(daemon,sizeof(daemon))) {
			int ret = system(daemon);
			if (0 != ret)
				warning("executing %s returned %d\n",daemon,ret);
		}
	}
	Endfunc = endwin;
	int source = setup_source();

	W = initscr();
	atexit((void (*)())endwin);
	start_color();
	nodelay(W, TRUE);
	init_pair(B_BLACK,COLOR_WHITE,COLOR_BLACK);
	init_pair(B_WHITE,COLOR_BLACK,COLOR_WHITE);
	init_pair(B_RED,COLOR_BLACK,COLOR_RED);
	init_pair(B_GREEN,COLOR_BLACK,COLOR_GREEN);
	init_pair(B_BLUE,COLOR_BLACK,COLOR_BLUE);
	init_pair(B_YELLOW,COLOR_BLACK,COLOR_YELLOW);
	init_pair(B_CYAN,COLOR_BLACK,COLOR_CYAN);
	init_pair(B_MAGENTA,COLOR_BLACK,COLOR_MAGENTA);
	init_pair(T_GREEN,COLOR_GREEN,COLOR_BLACK);
	init_pair(T_RED,COLOR_RED,COLOR_BLACK);
	init_pair(T_YELLOW,COLOR_YELLOW,COLOR_BLACK);
	init_pair(T_CYAN,COLOR_CYAN,COLOR_BLACK);
	getmaxyx(stdscr,Rows,Cols);
	attron(COLOR_PAIR(1));

	unsigned fl = fcntl(STDIN_FILENO,F_GETFL,0);
	if (-1 == fcntl(STDIN_FILENO,F_SETFL,fl | O_NONBLOCK)) 
		warning("could not set stdin to nonblocking: %s\n",strerror(errno));

	void (*osig)(int);
	osig = signal(SIGWINCH,sighandler);
	assert(osig != SIG_ERR);
	osig = signal(SIGINT,sighandler);
	assert(osig != SIG_ERR);

	for (;;) {
		const HostScope &t = process_input(source,1);

		if (Sigwinch) {
			struct winsize ws;
			if (-1 == ioctl(STDOUT_FILENO,TIOCGWINSZ,&ws))
				error("cannot read terminal size: %s\n",strerror(errno));
			Rows = ws.ws_row;
			Cols = ws.ws_col;
			dbug("SIGWINCH: Cols = %u, Rows = %u\n", Cols, Rows);
			if (Cols > 250)
				Cols = 250;
			endwin();
			W = initscr();
			start_color();
			refresh();
			redrawwin(stdscr);
			Sigwinch = false;
			(void) signal(SIGWINCH, sighandler);
		}
		clear();
		if (DisplayHelp) {
			display_help();
			continue;
		}

		unsigned ln = print_header(t);
		ln = print_cpus(ln,t);
		ln = print_mem(ln,t);
		ln = print_arc(ln,t);
		ln = print_users(ln,t);
		ln = print_disks(ln,t);
		ln = print_networks(ln,t);
		ln = print_procs(ln,t);
		int m = 0;
		while (ln < Rows-1) {
			const char *msg = get_msg(--m);
			if (msg == 0)
				break;
			if (0 == strncmp(msg,"warning",7))
				attron(COLOR_PAIR(T_YELLOW));
			else if (0 == strncmp(msg,"error",5))
				attron(COLOR_PAIR(T_RED));
			else
				attron(COLOR_PAIR(T_WHITE));
			mvaddstr(Rows-1+m,0,msg);
			++ln;
		}
		attron(COLOR_PAIR(T_CYAN));
		mvaddstr(Rows-1,0,"hosts:");
		attron(COLOR_PAIR(T_WHITE));
		for (unsigned i = 0; i < Hosts.size(); ++i) {
			if (i == Host)
				attron(A_BOLD);
			else
				attroff(A_BOLD);
			time_t t = time(0);
			if (RTime[i] == 0)
				attron(COLOR_PAIR(T_GREEN));
			else if (RTime[i] + 5 < t)
				attron(COLOR_PAIR(T_RED));
			else if (RTime[i] > t + 1)
				attron(COLOR_PAIR(T_YELLOW));
			else
				attron(COLOR_PAIR(T_WHITE));
			HostScope *hs = Hosts[i];
			assert(hs);
			const HostScope_HostInfo &hi = hs->hostinfo();
			const string &hn = hi.hostname();
			addch(' ');
			addstr(hn.c_str());
		}
		clrtoeol();
		attroff(A_BOLD);

		refresh();

	}
	endwin();
	return 0;
}
