/*
 *  Copyright (C) 2020, Thomas Maier-Komor
 *
 *  This is the source code of hostscope.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */


#include <arpa/inet.h>
#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#include <inttypes.h>
#include <netdb.h>
#include <signal.h>
#include <stdlib.h>
#include <string.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>

#include <string>
#include <iostream>
#include <sstream>

#ifdef WFC
#include "hostscope_wfc.h"
#else
#include "hostscope.pb.h"
#endif

#include "config.h"
#include "common.h"
#include "log.h"

#define LOCKFILE RUNDIR "/hostscopebr.pid"

using namespace std;

static bool ToStdout = false;
static struct sockaddr *UdpAddr = 0;
static size_t AddrLen = 0;
static int InfluxSock = -1;
static const char *InfluxHost = "localhost", *InfluxPort = "8090";


static void usage(const char *arg0, const char *m, int r)
{
	printf(
		"%s [option]\n"
		"Options:\n"
		"-h        : display this help screen and exit\n"
		"-I <h:p>  : send influx data to to <host>:<port>\n"
		"-f        : run in foreground mode - i.e. do not run as a daemon\n"
		"-c        : write Influx input in line protocol format to stdout\n"
		"-k        : kill hostscope's bridging daemon\n"
		"-l <log>  : write log to file <log>\n"
		"-m <a:p>  : set multicast group address and port (optional) to <a:p>\n"
		"-p <port> : set multicast group port to <port>\n"
		"-v <lvl>  : set verbose mode to level <lvl>\n"
		"-V        : display version number and exit\n"
		, arg0);
	if (m) 
		printf("\n%s\n",m);
	exit(r);
}


void setInfluxAddr(char *a)
{
	if (char *c = strchr(a,':')) {
		*c = 0;
		InfluxHost = strdup(a);
		InfluxPort = strdup(c+1);
	} else {
		InfluxPort = strdup(a);
	}
	info("influx interface set to %s:%s\n",InfluxHost,InfluxPort);
}


static int setup_source()
{
	struct sockaddr_in MA;
	struct ip_mreq mreq;
	int on = 1;
	int sock;

	sock = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
	if (-1 == sock)
		error("cannot create socket: %s\n",strerror(errno));
	if (-1 == setsockopt(sock,SOL_SOCKET,SO_REUSEADDR,&on,sizeof(on)))
		error("cannot set socket to reuse address: %s\n",strerror(errno));
	bzero(&MA,sizeof(MA));
	MA.sin_family = AF_INET;
	MA.sin_port = htons(Port);
	mreq.imr_interface.s_addr = htonl(INADDR_ANY);
	if (-1 == inet_pton(AF_INET, McastAddr, &MA.sin_addr))
		error("cannot set multicast address: %s\n",strerror(errno));
	memcpy(&mreq.imr_multiaddr.s_addr,(void *)&MA.sin_addr,sizeof(struct in_addr));
	if (-1 == bind(sock,(struct sockaddr *)&MA,sizeof(MA))) {
		error("cannot bind socket: %s\n",strerror(errno));
	} else if (-1 == setsockopt(sock,IPPROTO_IP,IP_ADD_MEMBERSHIP,&mreq,sizeof(mreq))) {
		error("could not join multicast group: %s\n",strerror(errno));
	} else
		return sock;
	close(sock);
	return -1;
}


static int receive_stats(int sock, char *buf, size_t bsize)
{
	int num;
	struct sockaddr_in from;
	socklen_t len = sizeof(from);
	num = recvfrom(sock,buf,bsize,0,(struct sockaddr *)&from,&len);
	if (num == -1)
		fatal("cannot read from socket: %s\n",strerror(errno));
	else 
		dbug("received %d bytes\n",num);
	return num;
}


static void parse_argv(const char *arg0, char a)
{
	switch (a) {
	default:
		abort();
		break;
	case 'c':
		ToStdout = true;
		Daemon = false;
		break;
	case 'f':
		Daemon = false;
		break;
	case 'I':
		setInfluxAddr(optarg);
		break;
	case 'h':
		usage(arg0,0,EXIT_SUCCESS);
		break;
	case 'k':
		kill_daemon(LOCKFILE);
		break;
	case 'l':
		setLogfile(optarg);
		break;
	case 'm':
		setMulticastAddr(optarg);
		break;
	case 'p':
		setMulticastPort(optarg);
		break;
	case 'v':
		setVerbose(optarg);
		break;
	case 'V':
		print_version();
		exit(EXIT_SUCCESS);
	case ':':
		usage(arg0,"missing argument",EXIT_FAILURE);
		break;
	case '?':
		usage(arg0,"unknown option",EXIT_FAILURE);
		break;
	}
}


static void toInflux(const HostScope &h, stringstream &ss)
{
	const HostScope_HostInfo &hi = h.hostinfo();
	const char *hostname = hi.hostname().c_str();

	const HostScope_LoadAvg &la = h.loadavg();
	double l1 = ((double)la.loadav1m())/1000.0;
	double l5 = ((double)la.loadav5m())/1000.0;
	double l15 = ((double)la.loadav15m())/1000.0;

	const HostScope_LinuxVmstat &v = h.linux_vmstat();
	const HostScope_Memory &mem = h.mem();
	uint64_t memfree = v.has_nr_free_pages() ?  v.nr_free_pages()<<2 : mem.free();
	memfree <<= 10;
	uint64_t pagedout = mem.pagedout() << 10;

	const HostScope_LinuxInfo &li = h.linux_info();

	ss	<< "hostscope,host=" << hostname
		<< " uptime=" << hi.uptime() 
		<< ",'lavg1m'=" << l1 << ",'lavg5m'=" << l5 << ",'lavg15m'=" << l15
		<< ",fork=" << li.forks() << ",ctxtsw=" << li.ctxtsw()
		<< ",intr=" << li.intrs() << ",sirqs=" << li.sirqs()
		<< ",memfree=" << memfree << ",pagedout=" << pagedout
		;
	if (v.has_pgpgin())
		ss << ",pgpgin=" << v.pgpgin();
	if (v.has_pgpgout())
		ss << ",pgpgout=" << v.pgpgout();
	if (v.has_pgfree())
		ss << ",pgfree=" << v.pgfree();
	if (v.has_pgfault())
		ss << ",pgfault=" << v.pgfault();

	if (h.linux_cpus_size()) {
		double usr = 0, sys = 0, irq = 0, idle = 0, n = 0;
	       	unsigned mhz = 0;
		for (const HostScope_LinuxCPU &cpu : h.linux_cpus()) {
			n += 1;
			idle += (double)cpu.idle()/100.0;
			irq += (double)cpu.irq()/100.0;
			usr += (double)cpu.user()/100.0;
			sys += (double)cpu.system()/100.0;
			mhz += cpu.currmhz();
		}
		ss << ",cpu_user=" << usr/n << ",cpu_sys=" << sys/n << ",cpu_irq=" << irq/n << ",cpu_idle=" << idle/n << ",mhz=" << mhz;
	}
	if (h.netdevices_size()) {
		uint64_t net_rx = 0, net_tx = 0;
		for (const HostScope_NetDevice &net : h.netdevices()) {
			net_rx += ((double)net.kreceived()) * 1024E-4;
			net_tx += ((double)net.ktransmit()) * 1024E-4;
		}
		ss << ",net_rxps=" << net_rx << ",net_txps=" << net_tx;
	}
	if (h.disks_size()) {
		uint64_t disk_rd = 0, disk_wr = 0;
		for (const HostScope_Disk &disk : h.disks()) {
			disk_rd += ((uint64_t)disk.kread()) << 10;
			disk_wr += ((uint64_t)disk.kwrite()) << 10;
		}
		ss << ",disk_rdps=" << disk_rd << ",disk_wrps=" << disk_wr;
	}
	ss << '\n';
}


static void sendStdout(const string &s)
{
	cout << "INSERT " << s << endl;
}


static void sendUdp(const string &s)
{
	if (-1 == sendto(InfluxSock,s.data(),s.size(),0,UdpAddr,AddrLen))
		error("error sending UDP packet: %s\n",strerror(errno));
}


static void initUdpTarget()
{
	struct addrinfo hint, *ret = 0;
	bzero(&hint,sizeof(hint));
	hint.ai_family = AF_UNSPEC;
	hint.ai_socktype = IPPROTO_UDP;
	hint.ai_socktype = SOCK_DGRAM;
	hint.ai_flags = AI_ADDRCONFIG;
	info("setting up UDP socket to %s:%s\n",InfluxHost,InfluxPort);
	int err = getaddrinfo(InfluxHost,InfluxPort,&hint,&ret);
	if ((err != 0) || (ret == 0))
		fatal("unable to resovle address information for '%s:%s': %s\n",InfluxHost,InfluxPort,strerror(errno));

	InfluxSock = socket(ret->ai_family,SOCK_DGRAM,IPPROTO_UDP);
	if (InfluxSock == -1)
		fatal("cannot create UDP socket: %s\n",strerror(errno));
	UdpAddr = ret->ai_addr;
	AddrLen = ret->ai_addrlen;
	//freeaddrinfo(ret);	! don't, still used
}


static void sigHandler(int sig)
{
	assert((sig == SIGINT) || (sig == SIGTERM));
	unlink(LOCKFILE);
	exit(EXIT_SUCCESS);
}


static void cleanup()
{
	unlink(LOCKFILE);
}


int main(int argc, char *argv[])
{
	int i;
#ifdef __linux
	signal(SIGSEGV,sigSegVhandler);
	signal(SIGABRT,sigSegVhandler);
#endif
	initSettings(argc,argv);
	while ((i = getopt(argc,argv,":cfhI:kl:m:p:Vv:")) != -1)
		parse_argv(argv[0],(char)i);
	if (Daemon) {
		start_daemon(LOCKFILE);
	} else {
		int pidf = create_lockfile(LOCKFILE);
		if (-1 == pidf)
			fatal("daemon already running\n");
		char buf[16];
		sprintf(buf,"%d\n",getpid());
		int n = write(pidf,buf,strlen(buf));
		if (n < 0)
			error("could not write pid to lockfile " LOCKFILE ": %s\n",strerror(errno));
		close(pidf);
	}
	signal(SIGTERM,sigHandler);
	signal(SIGINT,sigHandler);
	atexit(cleanup);
	if (!ToStdout)
		initUdpTarget();
	int sock = setup_source();
	char packet[UINT16_MAX];
	HostScope h;
	for (;;) {
		int n = receive_stats(sock,packet,sizeof(packet));
		if (n <= 0)
			continue;
		int x = h.fromMemory(packet,n);
		if (x != n) {
			warning("ignoring broken packet (%d/%u)\n",x,n);
			continue;
		}
		stringstream ss;
		toInflux(h,ss);
		if (ToStdout)
			sendStdout(ss.str());
		else
			sendUdp(ss.str());
		h.clear();
	}
}

