/*
 *  Copyright (C) 2015-2020 Thomas Maier-Komor
 *
 *  This is the source code of hostscope.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <dirent.h>
#include <errno.h>
#include <fcntl.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <unistd.h>

#ifdef WFC
#include "hostscope_wfc.h"
#else
#include "hostscope.pb.h"
#endif

#include "linuxstats.h"
#include "log.h"
#include "support.h"

#include <vector>

using namespace std;

struct CPU
{
	uint64_t timestamp;
	uint64_t user, nice, system, idle, iowait,
		 irq, softirq, steal, guest, guest_nice;
	int mhz_fd;

	CPU()
	: timestamp(0)
	, user(0), nice(0), system(0), idle(0), iowait(0)
	, irq(0), softirq(0), steal(0), guest(0), guest_nice(0)
	, mhz_fd(-2)
	{ }
};

struct Sensor
{
	string path;
	int fd;
	long id, crittemp;

	Sensor(const char *p, int f, long i, long c = 0)
	: path(p)
	, fd(f)
	, id(i)
	, crittemp(c)
	{ }
};

vector<CPU> Cpus;
vector<Sensor> Sensors;


static void read_currmhz(HostScope_LinuxCPU *cpu, int fd)
{
	TimeStat(__FUNCTION__);
	if (-1 == fd)
		return;
	char buf[256];
	int n = pread(fd,buf,sizeof(buf),0);
	if (n >= 0)
		cpu->set_currmhz(strtoll(buf,0,10)/1000);
}


static char *parse_cpu(HostScope *t, char *field, uint64_t ts, char *at)
{
	if (field[3] == 0) {
		dbug("skipping CPU summary\n");
		return at+3;
	}
	long id = strtol(field+3,0,10);
	if (id >= (long)Cpus.size()) 
		Cpus.resize(id+1);
	CPU *cpu = &Cpus[id];
	dbug("ts = %lu, cpu->timestamp = %lu\n",ts,cpu->timestamp);
	int nf = 0;
	char *af;
	uint64_t user = strtol(at,&af,10);
	if (at != af) ++nf;
	at = af;
	uint64_t nice = strtol(at,&af,10);
	if (at != af) ++nf;
	at = af;
	uint64_t system = strtol(at,&af,10);
	if (at != af) ++nf;
	at = af;
	uint64_t idle = strtol(at,&af,10);
	if (at != af) ++nf;
	at = af;
	uint64_t iowait = strtol(at,&af,10);
	if (at != af) ++nf;
	at = af;
	uint64_t irq = strtol(at,&af,10);
	if (at != af) ++nf;
	at = af;
	uint64_t softirq = strtol(at,&af,10);
	if (at != af) ++nf;
	at = af;
	uint64_t steal = strtol(at,&af,10);
	if (at != af) ++nf;
	at = af;
	uint64_t guest = strtol(at,&af,10);
	if (at != af) ++nf;
	at = af;
	uint64_t guest_nice = strtol(at,&af,10);
	if (at != af) ++nf;
	at = af;
	dbug("cpu %d: nf %d\n",id,nf);
	if (cpu->timestamp) {
		while (id >= (long)t->linux_cpus_size())
			t->add_linux_cpus();
		HostScope_LinuxCPU *tcpu = t->mutable_linux_cpus(id);
		assert(tcpu);
		tcpu->set_id(id);
		read_currmhz(tcpu,cpu->mhz_fd);
		double du = user-cpu->user;
		double di = idle-cpu->idle;
		double dn = nice-cpu->nice;
		double ds = system-cpu->system;
		double dw = iowait-cpu->iowait;
		double dq = irq-cpu->irq+softirq-cpu->softirq;
		double dv = steal-cpu->steal+guest-cpu->guest+guest_nice-cpu->guest_nice;
		double tt = du+di+dn+ds+dw+dq+dv;
		if (tt == 0) {
#ifndef WFC
			if (!tcpu->has_user())
				tcpu->set_user(0);
			if (!tcpu->has_nice())
				tcpu->set_user(0);
			if (!tcpu->has_system())
				tcpu->set_user(0);
			if (!tcpu->has_idle())
				tcpu->set_user(0);
#endif
			return at;
		}
		dbug("cpu %d du %g, user %lu, dt = %g\n",id,du,frac2fxptE4(du/tt),tt);
		dbug("cpu %d di %g, idle %lu, dt = %g\n",id,di,frac2fxptE4(di/tt),tt);
		dbug("cpu %d ds %g, system %lu, dt = %g\n",id,ds,frac2fxptE4(ds/tt),tt);
		tcpu->set_user(frac2fxptE4(du/tt));
		tcpu->set_nice(frac2fxptE4(dn/tt));
		tcpu->set_system(frac2fxptE4(ds/tt));
		tcpu->set_idle(frac2fxptE4(di/tt));
		if (nf >= 6)
			tcpu->set_iowait(frac2fxptE4(dw/tt));
		if (nf >= 7)
			tcpu->set_irq(frac2fxptE4(dq/tt));
		if (nf >= 9)
			tcpu->set_vhost(frac2fxptE4(dv/tt));
	} else {
		char filename[256];
		sprintf(filename,"/sys/bus/cpu/drivers/processor/cpu%lu/cpufreq/cpuinfo_cur_freq",id);
		cpu->mhz_fd = open(filename,O_RDONLY);
		if (cpu->mhz_fd == -1) {
			sprintf(filename,"/sys/devices/system/cpu/cpu%lu/cpufreq/cpuinfo_cur_freq",id);
			cpu->mhz_fd = open(filename,O_RDONLY);
		}
		if (cpu->mhz_fd == -1) {
			sprintf(filename,"/sys/bus/cpu/drivers/processor/cpu%lu/cpufreq/scaling_cur_freq",id);
			cpu->mhz_fd = open(filename,O_RDONLY);
		}
		if (cpu->mhz_fd == -1)
			warning("unable to find cpu frequency info files\n");
	}
	cpu->timestamp = ts;
	cpu->user = user;
	cpu->nice = nice;
	cpu->system = system;
	cpu->idle = idle;
	cpu->iowait = iowait;
	cpu->irq = irq;
	cpu->softirq = softirq;
	cpu->steal = steal;
	cpu->guest = guest;
	cpu->guest_nice = guest_nice;
	return at;
}


static void find_sensors(const char *path)
{
	TimeStat(__FUNCTION__);
	info("find_sensors('%s')\n",path);
	DIR *d = opendir(path);		// cannot be static, as multiple directories are searched...
	if (d == 0) {
		if (errno != ENOENT)
			warnonce("unable to search for sensors in %s: %s\n",path,strerror(errno));
		return;
	}
	char lfile[PATH_MAX];
	unsigned pl = strlen(path);
	strcpy(lfile,path);
	lfile[pl++] = '/';
	struct dirent *de;
	while ((de = readdir(d))) {
		if (de->d_name[0] == '.')
			continue;
		if (de->d_type == DT_DIR)
			continue;
		if (strncmp(de->d_name,"temp",4))
			continue;
		char *etype = 0;
		(void) strtol(de->d_name+4,&etype,10);
		if (etype == 0)
			continue;
		if (strncmp(etype,"_label",6))
			continue;
		strcpy(lfile+pl,de->d_name);
		int fd = open(lfile,O_RDONLY);
		if (fd == -1) {
			warnonce("unable to open temperature sensor label file %s: %s\n",lfile,strerror(errno));
			continue;
		}
		char buf[128];
		int n = read(fd,buf,sizeof(buf));
		close(fd);
		if (-1 == n) {
			warnonce("unable to read temperature sensor label file %s: %s\n",lfile,strerror(errno));
			continue;
		}
		if (0 == n) {
			warnonce("empty temperature sensor label file %s: %s\n",lfile,strerror(errno));
			continue;
		}
		if (buf[n-1] == '\n')
			buf[n-1] = 0;
		else
			buf[n] = 0;
		if (strncmp(buf,"Core ",5) != 0) {
			info("ignoring non-CPU temperature sensor %s\n",buf);
			continue;
		}
		long id = strtol(buf+5,0,10);
		char *at = strchr(lfile,'_');
		strcpy(at,"_input");
		info("adding temperature sensor '%s' at %s with id %d\n",buf,lfile,id);
		fd = open(lfile,O_RDONLY);
		if (fd == -1) 
			error("unable to open sensor file %s: %s\n",lfile,strerror(errno));
		else
			Sensors.push_back(Sensor(lfile,fd,id));
		id = strtol(buf+5,0,10);
		strcpy(etype,"_critical");
		fd = open(lfile,O_RDONLY);
		if (fd != -1) {
			n = read(fd,buf,sizeof(buf));
			close(fd);
			if (n > 0)
				Sensors.back().crittemp = strtol(buf,0,10);
			else
				warning("error reading from %s: %s\n",lfile,strerror(errno));
		}
	}
	closedir(d);
}

void init_tsensors()
{
	char path[PATH_MAX] = "/sys/class/hwmon/";
	DIR *d = opendir(path);
	if (d == 0)
		return;
	struct dirent *de;
	while ((de = readdir(d))) {
		if (de->d_name[0] == '.')
			continue;
		strcpy(path+17,de->d_name);
		find_sensors(path);
		strcat(path+17,"/device");
		find_sensors(path);
	}
	closedir(d);
}


void read_temperatures(HostScope *t)
{
	TimeStat(__FUNCTION__);
	for (unsigned i = 0; i < Sensors.size(); ++i) {
		Sensor &s = Sensors[i];
		char buf[64];
		buf[0] = 0;
		int n = pread(s.fd,buf,sizeof(buf),0);
		if (n < 0) 
			continue;
		long value = strtol(buf,0,10);
		value /= 100;
		dbug("sensor %s: %f degree (%s)\n",s.path.c_str(),(double)value/10.0,buf);
		if ((long)t->linux_cpus_size() > s.id) {
			HostScope_LinuxCPU *tcpu = t->mutable_linux_cpus(s.id);
			tcpu->set_currtemp(value);
			if (s.crittemp)
				tcpu->set_crittemp(s.crittemp/10.0);
			dbug("cpu %d: %3.1f degree\n",s.id,(double)value/10.0);
		}
	}
}


void read_cpuinfo(HostScope *t)
{
	TimeStat(__FUNCTION__);
	static uint64_t last_ts = 0;
	static uint64_t last_ctxt = 0;
	static uint64_t last_intr = 0;
	static uint64_t last_sirq = 0;
	static uint64_t last_procs = 0;
	static size_t bsize = 4096;
	static int sf = openat(ProcFd,"stat",O_RDONLY);
	char buf[bsize];
	char *b = buf;

#ifdef __CYGWIN__
	return;
#endif
	if (sf == -2)
		sf = openat(ProcFd,"stat",O_RDONLY);
	if (sf == -1) {
		error("unable to open /proc/stat: %s\n",strerror(errno));
		return;
	}
	uint64_t ts = get_us_time();
	int n = pread(sf,buf,sizeof(buf),0);
	if (n == -1) {
		close(sf);
		sf = -2;
		return;
	}
	while (n == (int)bsize) {
		if (-1 == lseek(sf,SEEK_SET,0)) {
			error("unable to seek /proc/stat to determine size");
			return;
		}
		size_t newsize = 0;
		do {
			n = read(sf,buf,sizeof(buf));
			if (-1 == n) {
				error("error scanning /proc/stat: %s\n",strerror(errno));
				return;
			}
			newsize += n;
		} while (n != 0);
		bsize = ((newsize/1024) + 1) * 1024;
		b = (char*) alloca(bsize);
		n = pread(sf,b,bsize,0);
		if (-1 == n) {
			error("error rereading /proc/stat: %s\n",strerror(errno));
			return;
		}
		dbug("bsize = %d\n",bsize);
	}
	if (n < 0) {
		error("unable to read /proc/stat: %s\n",strerror(errno));
		return;
	}
	double dt = (ts - last_ts) * 1E-6;
	b[n] = 0;
	HostScope_LinuxInfo *li = 0;
	if (last_intr)
		li = t->mutable_linux_info();
	char *at = b;
	while (at && *at) {
		char *field = at;
		at = strchr(at,' ');
		if (at == 0)
			return;
		*at = 0;
		++at;
		char *a = at;
		dbug("processing field %s\n",field);
		if (0 == strncmp(field,"cpu",3)) {
			at = parse_cpu(t,field,ts,at);
		} else if (0 == strcmp(field,"intr")) {
			uint64_t intr = 0;
			long long l = 0;
			do {
				at = a;
				intr += l;
				l = strtoll(at,&a,10);
			} while (at != a);
			if (last_intr != 0) 
				li->set_intrs(dbl2fxptE4((double)(intr-last_intr)/dt));
			dbug("intr %lu-%lu = %lu / %f = %f\n",intr,last_intr,intr-last_intr,dt,(double)(intr-last_intr)/dt);
			last_intr = intr;
		} else if (0 == strcmp(field,"ctxt")) {
			uint64_t ctxt = strtoll(at,&a,10);
			assert(at != a);
			at = a;
			dbug("ctxt %lu-%lu = %lu / %f = %f\n",ctxt,last_ctxt,ctxt-last_ctxt,dt,(double)(ctxt-last_ctxt)/dt);
			if (last_ctxt != 0)
				li->set_ctxtsw(dbl2fxptE4((double)(ctxt-last_ctxt)/dt));
			last_ctxt = ctxt;
		} else if (0 == strcmp(field,"processes")) {
			uint64_t procs = strtoll(at,&a,10);
			assert(at != a);
			at = a;
			dbug("procs %lu-%lu = %lu / %f = %f\n",procs,last_procs,procs-last_procs,dt,(double)(procs-last_procs)/dt);
			if (last_procs)
				li->set_forks(dbl2fxptE4((double)(procs-last_procs)/dt));
			last_procs = procs;
		} else if (0 == strcmp(field,"softirq")) {
			uint64_t sirq = 0;
			long long l = 0;
			char *a = at;
			do {
				at = a;
				sirq += l;
				l = strtoll(at,&a,10);
			} while (at != a);
			if (last_sirq != 0) 
				li->set_sirqs(dbl2fxptE4((double)(sirq-last_sirq)/dt));
			dbug("sirq %lu-%lu = %lu / %f = %f\n",sirq,last_sirq,sirq-last_sirq,dt,(double)(sirq-last_sirq)/dt);
			last_sirq = sirq;
		}
		//} else if (0 == strcmp(field,"btime")) {
		//} else if (0 == strcmp(field,"procs_running")) {
		//} else if (0 == strcmp(field,"procs_blocked")) {
		at = strchr(at,'\n');
		if (at)
			++at;
	}
	last_ts = ts;
	dbug("%d cpus\n",t->linux_cpus_size());
	read_temperatures(t);
}
