/*
 * Copyright 2015-2017, Thomas Maier-Komor
 * LICENSE: GPLv3, see file LICENSE for details
 */

#include <errno.h>
#include <fcntl.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#ifdef WFC
#include "hostscope_wfc.h"
#else
#include "hostscope.pb.h"
#endif


#include "log.h"
#include "support.h"

#ifndef O_BINARY
#define O_BINARY 0
#endif

void read_meminfo(HostScope *t)
{
	TimeStat(__FUNCTION__);
	static int mf = open("/proc/meminfo",O_RDONLY|O_BINARY);
	assert(mf != -1);
	char buf[4096];
	int n = pread(mf,buf,sizeof(buf),0);
	assert((n > 0) && (n < (int)sizeof(buf)));
	buf[n] = 0;
	HostScope_Memory *m = t->mutable_mem();
	uint64_t swaptotal = 0;
	char *at = buf;
	while (at && *at) {
		char *field = at;
		at = strchr(at,':');
		if (at == 0)
			return;
		*at = 0;
		++at;
		uint64_t value = strtol(at,&at,10);
		dbug("meminfo field '%s' value %lu\n",field,value);
		if (0 == strcmp(field,"MemTotal")) {
			dbug("setting memtotal to %lu\n",value);
			m->set_physical(value);
		} else if (0 == strcmp(field,"MemFree")) {
			m->set_free(value);
		} else if (0 == strcmp(field,"Mlocked")) {
			m->set_locked(value);
		} else if (0 == strcmp(field,"SwapTotal")) {
			m->set_swapsize(value);
			swaptotal = value;
		} else if (0 == strcmp(field,"SwapFree")) {
			m->set_pagedout(swaptotal - value);
		}
		at = strchr(at,'\n');
		if (at)
			++at;
	}
}


void read_vmstat(HostScope *t)
{
	TimeStat(__FUNCTION__);
	static int fd = open("/proc/vmstat",O_RDONLY);
	static uint64_t ts = 0, pgpgin = 0, pgpgout = 0, pgfree = 0, pgfault = 0;
	uint64_t now = get_us_time();
	double dt = now - ts;
	ts = now;
	if (fd == -1) {
		warning("unable to open /proc/vmstat: %s\n",strerror(errno));
		return;
	}
	char buf[4096];
	int n = pread(fd,buf,sizeof(buf),0);
	if (n < 0) {
		warning("unable to read /proc/vmstat: %s\n",strerror(errno));
		return;
	}
	assert(n < (int)sizeof(buf));
	buf[n] = 0;
	HostScope_LinuxVmstat &v = *t->mutable_linux_vmstat();
	char *at = buf;
	while (at && *at) {
		if (!strncmp(at,"nr_free_pages",13)) {
			long l = strtol(at+14,&at,10);
			if ((l != 0) || (errno == 0))
				v.set_nr_free_pages(l);
		} else if (!strncmp(at,"nr_mlock",8)) {
			long l = strtol(at+9,&at,10);
			if ((l != 0) || (errno == 0))
				v.set_nr_mlock(l);
		} else if (!strncmp(at,"nr_anon_pages",13)) {
			long l = strtol(at+14,&at,10);
			if ((l != 0) || (errno == 0))
				v.set_nr_anon_pages(l);
		} else if (!strncmp(at,"nr_file_pages",13)) {
			long l = strtol(at+14,&at,10);
			if ((l != 0) || (errno == 0))
				v.set_nr_file_pages(l);
		} else if (!strncmp(at,"nr_kernel_stack",15)) {
			long l = strtol(at+16,&at,10);
			if ((l != 0) || (errno == 0))
				v.set_nr_kernel_stack(l);
		} else if (!strncmp(at,"nr_shmem",8)) {
			long l = strtol(at+9,&at,10);
			if ((l != 0) || (errno == 0))
				v.set_nr_shmem(l);
		} else if (!strncmp(at,"pgpgin",6)) {
			long l = strtol(at+7,&at,10);
			if (l && pgpgin)
				v.set_pgpgin((double)((l-pgpgin)<<2)/dt);
			pgpgin = l;
		} else if (!strncmp(at,"pgpgout",7)) {
			long l = strtol(at+8,&at,10);
			if (l && pgpgout)
				v.set_pgpgout((double)((l-pgpgout)<<2)/dt);
			pgpgout = l;
		} else if (!strncmp(at,"pgfree",6)) {
			long l = strtol(at+7,&at,10);
			if (l && pgfree)
				v.set_pgfree((double)((l-pgfree)<<2)/dt);
			pgfree = l;
		} else if (!strncmp(at,"pgfault",7)) {
			long l = strtol(at+8,&at,10);
			if (l && pgfault)
				v.set_pgfault((double)((l-pgfault)<<2)/dt);
			pgfault = l;
		}
		at = strchr(at,'\n');
		if (at)
			++at;
	}
}
