/*
 *  Copyright (C) 2015-2021, Thomas Maier-Komor
 *
 *  This is the source code of hostscope.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <dirent.h>
#include <errno.h>
#include <fcntl.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <vector>

#ifdef WFC
#include "hostscope_wfc.h"
#else
#include "hostscope.pb.h"
#endif

#include "log.h"
#include "support.h"

using namespace std;

vector<string> PowerSupplies;


static void read_powerinfo(const char *ename, HostScope_PowerSupply &t)
{
	char fname[PATH_MAX];
	char *p = stpcpy(fname,"/sys/class/power_supply/");
	p = stpcpy(p,ename);
	p = stpcpy(p,"/uevent");
	int fd = open(fname,O_RDONLY);
	if (fd == -1) {
		warnonce("unable to open %s: %s\n",fname,strerror(errno));
		return;
	}
	char buf[4096];
	int n = read(fd,buf,sizeof(buf));
	close(fd);
	if (n < 0) {
		warnonce("unable to read %s: %s\n",fname,strerror(errno));
		return;
	}
	assert(n < (int)sizeof(buf));
	buf[n] = 0;
	char *at = buf;
	while (at && *at) {
		char *nl = strchr(at,'\n');
		assert(nl);
		*nl = 0;
		char *s = strchr(at,'=');
		if (s == 0) {
			at = nl + 1;
			continue;
		}
		*s = 0;
		if (0 == strcmp(at,"POWER_SUPPLY_NAME")) {
			at += 18;
			t.set_name(at);
			dbug("%s: name %s\n",fname,at);
		} else if (0 == strcmp(at,"POWER_SUPPLY_CAPACITY")) {
			at += 22;
			long capacity = strtol(at,0,10);
			t.set_capacity(capacity);
			dbug("%s: capacity %s\n",fname,at);
		} else if (0 == strcmp(at,"POWER_SUPPLY_ONLINE")) {
			at += 20;
			dbug("%s: state %s\n",fname,at);
			if (*at == '1')
				t.set_state("on");
			else if (*at == '0')
				t.set_state("off");
			else
				t.set_state("unknown");
		} else if (0 == strcmp(at,"POWER_SUPPLY_STATUS")) {
			at += 20;
			dbug("%s: state %s\n",fname,at);
			t.set_state(at);
		}
		at = nl + 1;
	}
}


void init_powersupplies(HostScope *t)
{
	TimeStat(__FUNCTION__);
	DIR *dir = opendir("/sys/class/power_supply");
	if (dir == 0) {
		warnonce("unable to open directory /sys/class/power_supply\n");
		return;
	}
	
	size_t id = 0;
	for (struct dirent *e = readdir(dir); e; e = readdir(dir)) {
		if ((e->d_name[0] == '.') || (0 == strncmp(e->d_name,"ucsi-source",11)))
			continue;
		while (id >= (size_t)t->power_supplies_size())
			t->add_power_supplies();
		HostScope_PowerSupply *p = t->mutable_power_supplies(id);
		read_powerinfo(e->d_name,*p);
		PowerSupplies.push_back(e->d_name);
		++id;
	}
	closedir(dir);
}


void read_powersupplies(HostScope *t)
{
	TimeStat(__FUNCTION__);
	for (size_t i = 0; i < (size_t)t->power_supplies_size(); ++i) {
		HostScope_PowerSupply *p = t->mutable_power_supplies(i);
		read_powerinfo(PowerSupplies[i].c_str(),*p);
	}
}

