/*
 *  Copyright (C) 2015-2020, Thomas Maier-Komor
 *
 *  This is the source code of hostscope.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <dirent.h>
#include <errno.h>
#include <fcntl.h>
#include <pwd.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <unistd.h>

#ifdef WFC
#include "hostscope_wfc.h"
#else
#include "hostscope.pb.h"
#endif

#include "log.h"
#include "support.h"

#include <map>
#include <string>


using namespace std;

extern bool ReadProcIO, ReadEuid, SendProcInfo, ReadStatm;
int MaxFd = 256;	// default value in case getrlimit fails
int ProcFd = -1;

static unsigned NumFds = 0;


struct UserInfo
{
	int32_t uid;
	uint32_t nproc, nthreads;
	int64_t rssmem,sharedmem,exclmem;
	double putime,pstime;
	double rchars,wchars;
	double minflts,majflts;
	string username;

	UserInfo()
	: uid(-1)
	, nproc(0)
	, nthreads(0)
	, rssmem(0)
	, sharedmem(-1)
//	, exclmem(-1)
	, putime(0)
	, pstime(0)
	, rchars(-1)
	, wchars(-1)
	, minflts(-1)
	, majflts(-1)
	, username()
	{ }
};

map<uint32_t,UserInfo> UserInfos;


struct ProcInfo
{
	uint32_t pid;
	int32_t uid;
	uint32_t nthreads;
	int stat_fd, statm_fd, io_fd;
	int64_t vsize,rssmem,sharedmem;
	uint64_t ts, minflt, majflt, utime, stime;
	uint64_t iots;
	int64_t rchar, wchar;
	double putime,pstime;
	double rchars,wchars;
	double minflts,majflts;
	string command;
	int keep;

	ProcInfo()
	: pid(0)
	, uid(-1)
	, nthreads(0)
	, stat_fd(-1)
	, statm_fd(-1)
	, io_fd(-1)
	, vsize(-1)
	, rssmem(-1)
	, sharedmem(-1)
	, ts(0)
	, minflt(0)
	, majflt(0)
	, rchar(-1)
	, wchar(-1)
	, putime(0)
	, pstime(0)
	, rchars(-1)
	, wchars(-1)
	, minflts(-1)
	, majflts(-1)
	, keep(0)
	{ }
};

map<uint32_t,ProcInfo> ProcInfos;


static inline char *skip_fields(unsigned u, char *at)
{
	do {
		at = strchr(at,' ');
		assert(at);
		++at;
	} while (--u);
	return at;
}


static void read_proc_statm(const char *pidstr, UserInfo *ui, ProcInfo *pi)
{
	TimeStat(__FUNCTION__);
	int fd;
	if (pi->statm_fd != -1) {
		fd = pi->statm_fd;
	} else {
		char pename[64];
		strcpy(stpcpy(pename,pidstr),"/statm");
		fd = openat(ProcFd,pename,O_RDONLY);
		if (fd == -1) {
			if (errno == ENFILE)
				error("file descriptor limit reached\n");
			return;
		}
		assert(MaxFd > 50);
	       	if (NumFds < (unsigned)(MaxFd - 50)) {
			++NumFds;
			pi->statm_fd = fd;
		}
	}
	if (fd == -1)
		return;
	char buf[4096];
	int n = pread(fd,buf,sizeof(buf),0);
	if (fd != pi->statm_fd)
		(void) close(fd);
	if (n <= 0) 
		return;
	assert(n < (int)sizeof(buf));
	buf[n] = 0;
	char *at = 0;
	if (pi->vsize == -1) {
		pi->vsize = strtoll(buf,&at,10);
		if (pi->vsize == 0)
			return;
	} else {
		// skip size
		at = strchr(buf,' ');
	}
	assert(at);
	at = strchr(at+1,' ');
	assert(at);
	char *c = 0;
	long long sharedmem = strtoll(at+1,&c,10);
	assert(c != at);
	pi->sharedmem = sharedmem;
	if (ui->sharedmem == -1)
		ui->sharedmem = sharedmem;
	else
		ui->sharedmem += sharedmem;
}


static UserInfo *read_proc_stat(const char *pidstr, ProcInfo *pi)
{
	TimeStat(__FUNCTION__);
	assert(pi);
	int fd = pi->stat_fd;
	if (fd == -1) {
		char pename[64];
		strcpy(stpcpy(pename,pidstr),"/stat");
		fd = openat(ProcFd,pename,O_RDONLY);
		if (fd == -1) {
			if (errno == ENFILE)
				error("file descriptor limit reached\n");
			return 0;
		}
		assert(MaxFd > 50);
	       	if (NumFds < (unsigned)(MaxFd - 50)) {
			++NumFds;
			pi->stat_fd = fd;
		}
	}
	char buf[4096];
	uint64_t ts = get_us_time();
	int n = pread(fd,buf,sizeof(buf),0);
	if (n < 0) {
		close(fd);
		pi->stat_fd = -1;
		return 0;
	}
	if (pi->stat_fd == -1)
		close(fd);
	assert(n < (int)sizeof(buf));
	buf[n] = 0;
	/*
		"%*u %*[^ ] %*c"	// pid, tcomm, state
		" %*u %*u %*u"		// ppid pgid sid
		" %*u %*u %*u"		// tty_nr tty_pgrp task->flags
		" %lu %*u %lu"		// min_flt cmin_flt maj_flt
		" %*u %lu %lu"		// cmaj_flt utime stime
		" %*u %*u %*u"		// cutime cstime priority
		" %*u %lu %*u"		// nice numthreads itvtime
		" %*u %*u %lu"		// start_time vsize rss
		" %*u %*u %*u"		// rsslim startcode endcode
		" %*u %*u %*u"		// startstack kstkesp kstkeip
		" %*u %*u %*u"		// signal blocked sigignore
		" %*u %*u %*u"		// sigcatch wchan nswap
	*/
	char *at = strchr(buf,')');	// tcomm is always in parantheses
	assert(at);
	at += 2;
	at = skip_fields(7,at);
	long long minflt = strtoll(at,&at,10);
	at = skip_fields(1,at+1);
	long long majflt = strtoll(at,&at,10);
	at = skip_fields(1,at+1);
	long long utime = strtoll(at,&at,10);
	long long stime = strtoll(at,&at,10);
	at = skip_fields(4,at+1);
	long long numthreads = strtoll(at,&at,10);
	at = skip_fields(3,at+1);
	long long rss = strtoll(at,&at,10);

	pi->nthreads = numthreads;
	UserInfo *ui = &UserInfos[pi->uid];
	++ui->nproc;
	ui->nthreads += numthreads;
	ui->rssmem += rss;
	pi->rssmem = rss;
	if (ui->username.empty()) {
		struct passwd *pwd = getpwuid(pi->uid);
		if (pwd)
			ui->username = pwd->pw_name;
		else {
			char username[64];
			sprintf(username,"<uid:%u>",pi->uid);
			ui->username = username;
		}
	}
	if (pi->ts != 0) {
		double dt = (double)(ts - pi->ts);
		pi->putime = (double)(utime-pi->utime)/dt;
		ui->putime += pi->putime;
		pi->pstime = (double)(stime-pi->stime)/dt;
		ui->pstime += pi->pstime;
		pi->minflts = (double)(minflt-pi->minflt)/dt;
		ui->minflts += pi->minflts;
		pi->majflts = (double)(majflt-pi->majflt)/dt;
		ui->majflts += pi->majflts;
	}
	pi->ts = ts;
	pi->utime = utime;
	pi->stime = stime;
	pi->minflt = minflt;
	pi->majflt = majflt;
	//info("process %u %s: %u threads, vss %lu, rss %lu\n",pid,comm, numthreads, vsize>>10, rss<<2);
	return ui;
}


static void read_proc_io(const char *pidstr, ProcInfo &pi, UserInfo &ui)
{
	TimeStat(__FUNCTION__);
	dbug("read_proc_io(%s)\n",pidstr);
	int fd;
	if (pi.io_fd != -1)
		fd = pi.io_fd;
	else {
		char pename[64];
		char *p = stpcpy(pename,pidstr);
		strcpy(p,"/io");
		fd = openat(ProcFd,pename,O_RDONLY);
		assert(MaxFd > 50);
	       	if (NumFds < (unsigned)(MaxFd - 50)) {
			++NumFds;
			pi.io_fd = fd;
		}
	}
	char buf[2048];
	uint64_t ts = get_us_time();
	if (fd == -1) {
		if (errno == ENFILE)
			error("file descriptor limit reached\n");
		return;
	}
	int n = pread(fd,buf,sizeof(buf),0);
	if (pi.io_fd == -1)
		close(fd);
	if (n < 0)
		return;
	assert(n < (int)sizeof(buf));
	buf[n] = 0;
	char *at = buf;
	while (at && *at) {
		if (0 == strncmp(at,"rchar:",6)) {
			long long rchar = strtoll(at+7,&at,10);
			if (ui.rchars < 0)	// yes, we gathered some data
				ui.rchars = 0;
			pi.rchars = (double)(rchar - pi.rchar)/(double)(ts-pi.iots);
			ui.rchars += pi.rchars;
			dbug("pid %s: delta rchar %lu\n",pidstr,rchar-pi.rchar);
			pi.rchar = rchar;
		} else if (0 == strncmp(at,"wchar:",6)) {
			long long wchar = strtoll(at+7,&at,10);
			if (ui.wchars < 0)	// yes, we gathered some data
				ui.wchars = 0;
			pi.wchars = (double)(wchar - pi.wchar)/(double)(ts-pi.iots);
			ui.wchars += pi.wchars;
			pi.wchar = wchar;
			break;
		}
		at = strchr(at,'\n');
		if (at)
			++at;
	}
	pi.iots = ts;
}


static void init_procinfo(const char *pidstr, ProcInfo *pi)
{
	TimeStat(__FUNCTION__);
	char commname[64];
	strcpy(stpcpy(commname,pidstr),"/comm");
	int cfd = openat(ProcFd,commname,O_RDONLY);
	char comm[PATH_MAX];
	int n = read(cfd,comm,sizeof(comm));
	close(cfd);
	if (n < 0)
		return;
	pi->command = string(comm,n);
	if (ReadEuid) {
		char pename[64];
		strcpy(stpcpy(pename,pidstr),"/status");
		char buf[2048];
		int fd = openat(ProcFd,pename,O_RDONLY);
		if (fd == -1)
			return;
		int n = read(fd,buf,sizeof(buf));
		close(fd);
		if (n < 0)
			return;
		assert(n < (int)sizeof(buf));
		buf[n] = 0;
		char *at = strstr(buf,"Uid:");
		assert(at);
		at += 5;
		unsigned uid = strtol(at,0,10);
		pi->uid = uid;
	} else {
		struct stat st;
		int err = fstatat(ProcFd,pidstr,&st,0);
		if (err != -1)
			pi->uid = st.st_uid;
	}
	pi->pid = strtol(pidstr,0,10);
}


void read_procs(HostScope *t)
{
	TimeStat(__FUNCTION__);
	static DIR *dir = opendir("/proc");
	static int euid = geteuid();

	for (map<uint32_t,UserInfo>::iterator i(UserInfos.begin()),e(UserInfos.end()); i != e; ++i) {
		i->second.nproc = 0;
		i->second.nthreads = 0;
		i->second.rssmem = -1;
//		i->second.exclmem = -1;
		i->second.sharedmem = -1;
		i->second.minflts = 0;
		i->second.majflts = 0;
		i->second.putime = 0;
		i->second.pstime = 0;
		i->second.rchars = -1;
		i->second.wchars = -1;
	}
	uint64_t start_ts = get_us_time();
	if (dir == 0) {
		error("unable to open /proc dir: %s\n",strerror(errno));
		return;
	}
	for (struct dirent *de = readdir(dir); de ; de = readdir(dir)) {
		if ((de->d_name[0] < '0') || (de->d_name[0] > '9'))
			continue;
		long pid = strtol(de->d_name,0,10);
		if (0 == pid)
			continue;
		ProcInfo *pi = &ProcInfos[pid];
		if (pi->uid == -1) 
			init_procinfo(de->d_name,pi);
		UserInfo *ui = read_proc_stat(de->d_name,pi);	// for utime/stime
		if (ui == 0)
			continue;
		if (pi->vsize && ReadStatm)
			read_proc_statm(de->d_name,ui,pi);
		if (ReadProcIO && ((euid == 0) || (euid == pi->uid)))		// rchar/wchar need setuid root
			read_proc_io(de->d_name,*pi,*ui);
	}
	rewinddir(dir);
	
	/* remove processes that no longer exist */
	uint32_t killed = 0;
	map<uint32_t,ProcInfo>::iterator i = ProcInfos.begin();
	while (i != ProcInfos.end()) {
		map<uint32_t,ProcInfo>::iterator d = i;
		++i;
		if (d->second.ts < start_ts) {
			if (d->second.stat_fd != -1) {
				close(d->second.stat_fd);
				--NumFds;
			}
			if (d->second.statm_fd != -1) {
				close(d->second.statm_fd);
				--NumFds;
			}
			if (d->second.io_fd != -1) {
				close(d->second.io_fd);
				--NumFds;
			}
			ProcInfos.erase(d);
			++killed;
		}
	}

	for (map<uint32_t,UserInfo>::iterator i(UserInfos.begin()),e(UserInfos.end()); i != e; ++i) {
		dbug("%s: nproc = %u, nthread = %u, rss = %lu\n"
				, i->second.username.c_str()
				, i->second.nproc
				, i->second.nthreads
				, i->second.rssmem);
		unsigned nu = t->users_size();
		HostScope_User *tu = 0;
	       	if (nu)
			tu = t->mutable_users(0);
		UserInfo &u = i->second;
		unsigned idx = 0;
		while ((idx < nu) && (tu != 0) && (u.username != tu->name())) {
			tu = t->mutable_users(idx);
			++idx;
		} 
		if ((tu == 0) || (u.username != tu->name())) {
			tu = t->add_users();
			tu->set_name(u.username);
			tu->set_uid(i->first);
		}
		tu->set_nproc(u.nproc);
		if (u.nthreads > 0)
			tu->set_nthreads(u.nthreads);
		if (u.rssmem != -1)
			tu->set_rssmem(u.rssmem<<2);
//		if (u.exclmem != -1)
//			tu->set_exclmem(u.exclmem<<2);
		if (u.sharedmem != -1)
			tu->set_sharedmem(u.sharedmem<<2);
		tu->set_minflts(dbl2fxptE4(u.minflts*1E6));
		tu->set_majflts(dbl2fxptE4(u.majflts*1E6));
		tu->set_putime(dbl2fxptE4(u.putime*1E4));	// can be > 100%, so use dbl2fxptE4 instead of frac2fxptE4
		tu->set_pstime(dbl2fxptE4(u.pstime*1E4));	// can be > 100%, so use dbl2fxptE4 instead of frac2fxptE4
		if (u.rchars >= 0)
			tu->set_rchars(dbl2fxptE4(u.rchars));
		if (u.wchars >= 0)
			tu->set_wchars(dbl2fxptE4(u.wchars));
	}
	if (!SendProcInfo)
		return;
	t->clear_processes();
	for (map<uint32_t,ProcInfo>::iterator i(ProcInfos.begin()),e(ProcInfos.end()); i != e; ++i) {
		ProcInfo &p = i->second;
		if ((p.minflts < 1E-4) && (p.majflts < 1E-4) && (p.putime < 1E-6) && (p.pstime < 1E-6) && (p.rchars < 1E3) && (p.wchars < 1E3)) {
			if (p.keep == 0)
				continue;
			--p.keep;
		} else
			p.keep = 3;
		HostScope_Process *tp = t->add_processes();
		tp->set_pid(p.pid);
		tp->set_uid(p.uid);
		tp->set_command(p.command);
		if (p.nthreads > 0)
			tp->set_nthreads(p.nthreads);
		if (p.rssmem != -1)
			tp->set_rssmem(p.rssmem<<2);
		if (p.sharedmem != -1)
			tp->set_sharedmem(p.sharedmem<<2);
		tp->set_minflts(dbl2fxptE4(p.minflts*1E6));
		tp->set_majflts(dbl2fxptE4(p.majflts*1E6));
		tp->set_putime(dbl2fxptE4(p.putime*1E4));
		tp->set_pstime(dbl2fxptE4(p.pstime*1E4));
		if (p.rchars >= 0.0)
			tp->set_rchars(dbl2fxptE4(p.rchars));
		if (p.wchars >= 0.0)
			tp->set_wchars(dbl2fxptE4(p.wchars));
	}
}
