/*
 *  Copyright (C) 2015-2020, Thomas Maier-Komor
 *
 *  This is the source code of hostscope.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <string.h>
#include <syslog.h>
#include <unistd.h>

#include <set>
#include <string>
#include <vector>

#include "log.h"

using namespace std;

#ifdef __GNUC__
extern FILE *popen(const char *,const char *);
extern int pclose(FILE *);
extern int snprintf(char *,size_t,const char *,...);
extern int vsnprintf(char *,size_t,const char *,va_list);
#endif

char **ARGV;
int ARGC;
void (*DfltSegV)(int signr) = 0;
int Syslog = 0, Verbose = 3, Log = -1;
unsigned History = 0;

vector<string> Msgs;

static void add_msg(const char *msg)
{
	Msgs.push_back(msg);
	if (Msgs.size() > History)
		Msgs.erase(Msgs.begin());
}


const char *get_msg(int i)
{
	int nm = Msgs.size();
	dbug("get_msg(%d), nm %u, nm+i=%d\n",i,nm,nm+i);
	if (nm == 0)
		return 0;
	if (i < 0) {
		if (-i > nm)
			return 0;
		return Msgs[nm+i].c_str();
	}
	if (i >= nm)
		return 0;
	return Msgs[i].c_str();
}


#ifdef DEBUG
void dbug(const char *msg, ...)
{
	if (Verbose < 5)
		return;
	va_list val;
	char buf[8192];

	va_start(val,msg);
	if (Syslog) {
		vsyslog(LOG_WARNING,msg,val);
	} else {
		size_t l = vsnprintf(buf,sizeof(buf),msg,val);
		if (-1 == write(Log == -1 ? STDERR_FILENO : Log, buf, l))
			syslog(LOG_INFO,"%s",buf+6);
	}
	va_end(val);
}
#endif


void info(const char *msg, ...)
{
	if (Verbose < 4)
		return;
	va_list val;
	char buf[8192] = "info: ";
	size_t l = 6;

	va_start(val,msg);
	l += vsnprintf(buf+l,sizeof(buf)-l,msg,val);
	va_end(val);
	if (History)
		add_msg(buf);
	if (Syslog) {
		syslog(LOG_INFO,"%s",buf+6);
	} else if (-1 == write(Log == -1 ? STDERR_FILENO : Log, buf, l)) {
		if (Syslog)
			syslog(LOG_WARNING,"error writing to log file: %s",strerror(errno));
	}
}


void warnonce(const char *msg, ...)
{
	static set<string> warnings;
	va_list val;
	char buf[8192] = "warning: ";
	size_t l = 9;

	va_start(val,msg);
	l += vsnprintf(buf+9,sizeof(buf)-9,msg,val);
	va_end(val);
	if (!warnings.insert((string)(buf+9)).second)
		return;
	if (History)
		add_msg(buf);
	if (Syslog) {
		syslog(LOG_WARNING,"%s",buf+9);
	} else if (-1 == write(Log == -1 ? STDERR_FILENO : Log, buf, l)) {
		if (Syslog)
			syslog(LOG_WARNING,"error writing to log file: %s",strerror(errno));
	}
}


void warning(const char *msg, ...)
{
	va_list val;
	char buf[8192] = "warning: ";
	size_t l = 9;

	va_start(val,msg);
	l += vsnprintf(buf+9,sizeof(buf)-9,msg,val);
	va_end(val);
	if (History)
		add_msg(buf);
	if (Syslog) {
		syslog(LOG_WARNING,"%s",buf+9);
	} else if (-1 == write(Log == -1 ? STDERR_FILENO : Log, buf, l)) {
		if (Syslog)
			syslog(LOG_WARNING,"error writing to log file: %s",strerror(errno));
	}
}


void error(const char *msg, ...)
{
	va_list val;
	char buf[8192] = "error: ";
	size_t l = 7;

	va_start(val,msg);
	l += vsnprintf(buf+7,sizeof(buf)-7,msg,val);
	va_end(val);
	if (History)
		add_msg(buf);
	if (Syslog)
		syslog(LOG_ERR,"%s",buf+7);
	if (Log != -1) {
		if ((-1 == write(Log, buf, l)) && (Syslog))
			syslog(LOG_WARNING,"error writing to log file: %s",strerror(errno));
	}
	if ((-1 == write(STDERR_FILENO, buf, l)) && (Syslog))
		syslog(LOG_WARNING,"error writing to stderr: %s",strerror(errno));
}


void fatal(const char *msg, ...)
{
	va_list val;
	char buf[8192] = "fatal: ";
	size_t l = 7;

	va_start(val,msg);
	l += vsnprintf(buf+7,sizeof(buf)-7,msg,val);
	va_end(val);
	if (Syslog)
		vsyslog(LOG_ERR,msg,val);
	if (Log != -1) {
		if (-1 == write(Log, buf, l))
			(void) close(Log);
	}
	if (-1 == write(STDERR_FILENO, buf, l))
		(void) close(STDERR_FILENO);
	exit(EXIT_FAILURE);
}


void log2syslog(const char *exename)
{
	openlog(exename,LOG_USER,0);
	Syslog = 1;
}


void setLogfile(const char *logfile)
{
	info("switching to logfile %s\n",logfile);
	int log = open(logfile,O_WRONLY|O_APPEND|O_CREAT|O_TRUNC,0666);
	if (log == -1) {
		warning("could not create logfile %s: %s\n",logfile,strerror(errno));
	} else {
		(void) close(Log);
		Log = log;
	}
}


void setVerbose(const char *valuestr)
{
	char *eptr;
	long v = strtol(valuestr,&eptr,0);
	if (*eptr) {
		error("invalid argument '%s' for option verbose\n",valuestr);
	} else	if ((v <= 0) || (v > 5)) {
		warning("verbosity level '%d' out of range - value ignored\n",v);
	} else {
		Verbose = v;
		info("set verbose to %d\n",v);
	}
}


