/*
 *  Copyright (C) 2015-2020, Thomas Maier-Komor
 *
 *  This is the source code of hostscope.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef SUPPORT_H
#define SUPPORT_H

#include <assert.h>
#include <time.h>
#include <math.h>
#include <stdio.h>
#include <string.h>
#include <stdint.h>


#ifdef PROFILE
#include "rta.h"
#else
#define TimeStat(x)
#endif


inline uint64_t get_us_time()
{
	struct timespec ts;
	clock_gettime(CLOCK_MONOTONIC,&ts);
	return ts.tv_sec * 1000000LL + ts.tv_nsec / 1000;
}


inline uint32_t frac2fxptE4(double f)
{
	if (isnan(f))
		return 0;
	assert((f >= 0) && (f <= 1));
	return (uint32_t) nearbyint(f*1E4);
}


inline uint32_t dbl2fxptE4(double f)
{
	return (uint32_t) nearbyint(f*1E4);
}



static inline char *
kb2str_r(double d, char *s)
{
	const char *dim = "kMGT", *f;

	double v = d;
	while (v >= 10000.0) {
		v /= 1024.0;
		++dim;
		if (*dim == 0) {
			v *= 1024.0*1024.0*1024.0*1024.0;
			break;
		}
	}
	if (v < 0)
		f = "<err> ";
	else if (rint(v) < 100.0)
		f = "%4.1f%ci";
	else if (rint(v) < 10000.0)
		f = "%4.0f%ci";
	else
		f = "%6.g";
	return s + sprintf(s,f,v,*dim);
}


/*
static inline char *
kb2str(double v)
{
	static char str[8];
	kb2str_r(v,str);
	return str;
}
*/


static inline char *
word2perc(uint32_t u, char *s)
{
	double d = (double)u;
	d /= 100;
	u = rint(d);
	return s + sprintf(s,"%3u",u);
}


/*
static inline char *
per2str_r(double d, char *s)
{
	const char *f;
	assert(d <= 1E2);
	if (d < 0) {
		f = "n/a ";
	} else if (d < 2E-2) {
		f = "%4.0f";
	} else if (d < 9.995) {
		f = "%4.2f";
	} else if (d < 99.95) {
		f = "%4.1f";
	} else {
		f = "%4.0f";
		d = rint(d);
	}
	(void) sprintf(s,f,d);
	return s;
}


static inline char *
per2str(double d)
{
	static char str[8];
	const char *f;
	assert(d <= 1E2);
	if (d < 0) {
		f = "n/a ";
	} else if (d < 2E-2) {
		f = "%4.0f";
	} else if (d < 9.995) {
		f = "%4.2f";
	} else if (d < 99.95) {
		f = "%4.1f";
	} else {
		f = "%4.0f";
		d = rint(d);
	}
	(void) sprintf(str,f,d);
	return str;
}


static inline char *
dbl2str3_r(double d, char *s)
{
	const char *dim = " kMGT", *f;
	double r = d;

	while (r >= 1E4) {
		r /= 1E3;
		++dim;
		if (*dim == 0) {
			r *= 1E15;
			d = -1;
			break;
		}
	}
	if (*dim == ' ') {
		if (r < 1E-2)
			f = "%4.0f";
		else if (r < 9.995)
			f = "%4.2f";
		else if (r < 1E2)
			f = "%4.1f";
		else if (r < 1E4)
			f = "%4.0f";
		else
			assert(0);
	} else if (r < 1E-2)
		f = "%3.0f%c";
	else if (r < 9.995)
		f = "%3.1f%c";
	else if (r < 1E4)
		f = "%3.0f%c";
	else
		f = "%4.g";
	(void) sprintf(s,f,r,*dim);
	return s;
}
*/


static inline char *
dbl2str_r(double d, char *s)
{
	const char *dim = " kMGT", *f;
	double r = d;

	if ((d >= 0.99) && (d < 1))
		d = 1;
	while (r >= 1E4) {
		r /= 1E3;
		++dim;
		if (*dim == 0) {
			r *= 1E15;
			d = -1;
			break;
		}
	}
	if (d < 0)
		f = " unav.";
	else if (r < 2E-2)
		f = "%4.0f%c";
	else if (r < 9.995)
		f = "%4.2f%c";
	else if (r < 1E2)
		f = "%4.1f%c";
	else if (r < 1E4)
		f = "%4.0f%c";
	else
		f = "%5.g";
	return s + sprintf(s,f,r,*dim);
}

/*
static inline char *
dbl2str(double d)
{
	static char str[8];
	return dbl2str_r(d,str);
}
*/


#ifdef __CYGWIN__
static inline int pread_cygw(int fd, char *buf, size_t s, off_t o)
{
	int r = lseek(fd,o,SEEK_SET);
	if (r == -1)
		return -1;
	return read(fd,buf,s);
}
#define pread pread_cygw
#endif

#endif
