/*
 *  Copyright (C) 2020 Thomas Maier-Komor
 *
 *  This is the source code of hostscope.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <dirent.h>
#include <errno.h>
#include <fcntl.h>
#include <inttypes.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>

#ifdef WFC
#include "hostscope_wfc.h"
#else
#include "hostscope.pb.h"
#endif

#include "linuxstats.h"
#include "log.h"
#include "support.h"


struct ARC
{
	uint64_t hits, misses, d_hit, d_miss, m_hit, m_miss, pf_hit, pf_miss,
		 l2_read, l2_write, l2_hit, l2_miss, l2_size
		;
};


void read_arc(HostScope *t)
{
	static int fd = -2;
	static ARC larc = {0};
	static uint64_t last_ts;

	if (fd == -2)
		fd = open("/proc/spl/kstat/zfs/arcstats",O_RDONLY);
	if (fd == -1)
		return;

	uint64_t ts = get_us_time();
	double dt = (ts - last_ts) * 1E-6;
	last_ts = ts;
	char buf[8192];
	int n = pread(fd,buf,sizeof(buf)-1,0);
	if (n == -1) {
		close(fd);
		fd = -2;
		return;
	}
	buf[n] = 0;
	HostScope_ARC *marc = t->mutable_arc();
	marc->clear();

	ARC arc;
	if (const char *sizestr = strstr(buf,"size ")) {
		uint64_t size;
		if (1 == sscanf(sizestr+37,"%" SCNu64,&size))
			marc->set_size(size>>10);
	}

	if (const char *l2sizestr = strstr(buf,"l2_size ")) {
		uint64_t l2_size;
		if (1 == sscanf(l2sizestr+37,"%" SCNu64,&l2_size))
			marc->set_l2size(l2_size>>10);
	}

	if (const char *l2hitstr = strstr(buf,"l2_hits ")) {
		if (1 == sscanf(l2hitstr+37,"%" SCNu64,&arc.l2_hit))
			marc->set_l2_hit((double)(arc.l2_hit-larc.l2_hit)/dt);
	}

	if (const char *l2missstr = strstr(buf,"l2_misses ")) {
		if (1 == sscanf(l2missstr+37,"%" SCNu64,&arc.l2_miss))
			marc->set_l2_miss((double)(arc.l2_miss-larc.l2_miss)/dt);
	}

	if (const char *dhitstr = strstr(buf,"demand_data_hits ")) {
		if (1 == sscanf(dhitstr+37,"%" SCNu64,&arc.d_hit))
			marc->set_data_hit(arc.d_hit-larc.d_hit);
	}

	if (const char *dmissstr = strstr(buf,"demand_data_misses ")) {
		if (1 == sscanf(dmissstr+37,"%" SCNu64,&arc.d_miss))
			marc->set_data_miss(arc.d_miss-larc.d_miss);
	}

	if (const char *mhitstr = strstr(buf,"demand_metadata_hits ")) {
		if (1 == sscanf(mhitstr+37,"%" SCNu64,&arc.m_hit))
			marc->set_metadata_hit(arc.m_hit-larc.m_hit);
	}

	if (const char *mmissstr = strstr(buf,"demand_metadata_misses ")) {
		if (1 == sscanf(mmissstr+37,"%" SCNu64,&arc.m_miss))
			marc->set_metadata_miss(arc.m_miss-larc.m_miss);
	}

	if (const char *phitstr = strstr(buf,"prefetch_data_hits ")) {
		uint64_t pfhit;
		if (1 == sscanf(phitstr+37,"%" SCNu64,&pfhit))
			arc.pf_hit = pfhit;
	}
	if (const char *phitstr = strstr(buf,"prefetch_metadata_hits ")) {
		uint64_t pfhit;
		if (1 == sscanf(phitstr+37,"%" SCNu64,&pfhit))
			arc.pf_hit += pfhit;
	}
	marc->set_prefetch_hit(arc.pf_hit-larc.pf_hit);

	if (const char *pmissstr = strstr(buf,"prefetch_data_misses ")) {
		uint64_t pfmiss;
		if (1 == sscanf(pmissstr+37,"%" SCNu64,&pfmiss))
			arc.pf_miss = pfmiss;
	}
	if (const char *pmissstr = strstr(buf,"prefetch_metadata_misses ")) {
		uint64_t pfmiss;
		if (1 == sscanf(pmissstr+37,"%" SCNu64,&pfmiss))
			arc.pf_miss += pfmiss;
	}
	marc->set_prefetch_miss(arc.pf_miss-larc.pf_miss);

	if (const char *l2hstr = strstr(buf,"l2_hits ")) {
		if (1 == sscanf(l2hstr+37,"%" SCNu64,&arc.l2_hit))
			marc->set_l2_hit(arc.l2_hit-larc.l2_hit);
	}

	if (const char *l2mstr = strstr(buf,"l2_misses ")) {
		if (1 == sscanf(l2mstr+37,"%" SCNu64,&arc.l2_miss))
			marc->set_l2_hit(arc.l2_miss-larc.l2_miss);
	}

	if (const char *l2rstr = strstr(buf,"l2_read_bytes ")) {
		if (1 == sscanf(l2rstr+37,"%" SCNu64,&arc.l2_read))
			marc->set_l2_read((arc.l2_read-larc.l2_read) >> 10);
	}

	if (const char *l2rstr = strstr(buf,"l2_write_bytes ")) {
		if (1 == sscanf(l2rstr+37,"%" SCNu64,&arc.l2_write))
			marc->set_l2_write((arc.l2_write-larc.l2_write) >> 10);
	}

	larc = arc;
}
